import scipy as sp
import numpy as np


def rbf_kernel(X, Z, variance, lengthscale):
    """
    A classic radial-basis function (Gaussian; exponential squared) covariance kernel

    .. math::
        \\kappa(X, Z | \\sigma^2, \\Lambda) = \\sigma^2 \\exp\\left[-\\frac12 (X-Z)^T \\Lambda^{-1} (X-Z) \\right]

    Parameters
    ----------
    X : np.ndarray
        The first set of points
    Z : np.ndarray
        The second set of points. Must have same second dimension as `X`
    variance : double
        The amplitude for the RBF kernel
    lengthscale : np.ndarray or double
        The lengthscale for the RBF kernel. Must have same second dimension as `X`

    """

    sq_dist = sp.spatial.distance.cdist(X / lengthscale, Z / lengthscale, "sqeuclidean")
    return variance * np.exp(-0.5 * sq_dist)


def batch_kernel(X, Z, variances, lengthscales):
    """
    Batched RBF kernel

    Parameters
    ----------
    X : np.ndarray
        The first set of points
    Z : np.ndarray
        The second set of points. Must have same second dimension as `X`
    variances : np.ndarray
        The amplitude for the RBF kernel
    lengthscales : np.ndarray
        The lengthscale for the RBF kernel. Must have same second dimension as `X`

    See Also
    --------
    :function:`rbf_kernel`
    """
    blocks = [rbf_kernel(X, Z, var, ls) for var, ls in zip(variances, lengthscales)]
    return sp.linalg.block_diag(*blocks)



from .instruments import *
from .base_interfaces import *
from .interfaces import *
from .interpolators import *
from .utils import *

__all__ = [s for s in dir() if not s.startswith("_")]  # Remove dunders.



import bz2
import logging
import os

import numpy as np
from astropy.io import fits, ascii
from scipy.interpolate import InterpolatedUnivariateSpline

import Starfish.constants as C
from Starfish.utils import create_log_lam_grid
from .base_interfaces import GridInterface
from .utils import vacuum_to_air, idl_float

log = logging.getLogger(__name__)


class PHOENIXGridInterface(GridInterface):
    """
    An Interface to the PHOENIX/Husser synthetic library.

    Note that the wavelengths in the spectra are in Angstrom and the flux are in :math:`F_\\lambda` as
    :math:`erg/s/cm^2/cm`

    Parameters
    ----------
    path : str or path-like
        The path of the base of the PHOENIX library
    air : bool, optional
        Whether the wavelengths are measured in air or not. Default is True
    wl_range : tuple, optional
        The (min, max) of the wavelengths, in :math:`\\AA`. Default is (3000, 54000), which is the full wavelength grid for PHOENIX.
    """

    def __init__(self, path, air=True, wl_range=(3000, 54000)):
        super().__init__(
            name="PHOENIX",
            param_names=["T", "logg", "Z", "alpha"],
            points=[
                np.hstack([np.arange(2300, 7000, 100), np.arange(7000, 12001, 200)]),
                np.arange(0.5, 6.1, 0.5),
                np.hstack([np.arange(-4.0, -2, 1), np.arange(-2, 1.1, 0.5)]),
                np.arange(-0.2, 1.21, 0.2),
            ],
            wave_units="AA",
            flux_units="erg/s/cm^2/cm",
            air=air,
            wl_range=wl_range,
            path=path,
        )  # wl_range used to be (2999, 13001)

        # These dictionaries provide the link between numerical representation and the
        # string representation for Z and alpha
        self.par_dicts = [
            None,
            None,
            {
                -2: "-2.0",
                -1.5: "-1.5",
                -1: "-1.0",
                -0.5: "-0.5",
                0.0: "-0.0",
                0.5: "+0.5",
                1: "+1.0",
            },
            {
                -0.4: ".Alpha=-0.40",
                -0.2: ".Alpha=-0.20",
                0.0: "",
                0.2: ".Alpha=+0.20",
                0.4: ".Alpha=+0.40",
                0.6: ".Alpha=+0.60",
                0.8: ".Alpha=+0.80",
            },
        ]

        try:
            wl_filename = os.path.join(
                self.path, "WAVE_PHOENIX-ACES-AGSS-COND-2011.fits"
            )
            w_full = fits.getdata(wl_filename)
        except:
            raise ValueError("Wavelength file improperly specified.")

        # if air is true, convert the normally vacuum file to air wls.
        if self.air:
            self.wl_full = vacuum_to_air(w_full)
        else:
            self.wl_full = w_full

        self.ind = (self.wl_full >= self.wl_range[0]) & (
            self.wl_full <= self.wl_range[1]
        )
        self.wl = self.wl_full[self.ind]
        self.rname = "Z{2:}{3:}/lte{0:0>5.0f}-{1:.2f}{2:}{3:}.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits"
        self.full_rname = os.path.join(self.path, self.rname)

    def check_params(self, parameters):
        super().check_params(parameters)

        if parameters[3] != 0:
            if (parameters[0] < 3500) or (parameters[0] > 8000):
                raise ValueError(
                    "Alpha concentration does not run for T={}".format(parameters[0])
                )
            if (parameters[2] < -3.0) or (parameters[2] > 0):
                raise ValueError(
                    "Alpha concentration does not run for Z={}".format(parameters[2])
                )

        return True

    def load_flux(self, parameters, header=False, norm=True):
        self.check_params(parameters)  # Check to make sure that the keys are
        # allowed and that the values are in the grid

        # Create a list of the parameters to be fed to the format string
        # optionally replacing arguments using the dictionaries, if the formatting
        # of a certain parameter is tricky
        str_parameters = []
        for param, par_dict in zip(parameters, self.par_dicts):
            if par_dict is None:
                str_parameters.append(param)
            else:
                str_parameters.append(par_dict[param])

        fname = self.full_rname.format(*str_parameters)

        # Still need to check that file is in the grid, otherwise raise a C.GridError
        # Read all metadata in from the FITS header, and append to spectrum
        if not os.path.exists(fname):
            raise ValueError("{} is not on disk.".format(fname))

        hdu_list = fits.open(fname)
        flux = hdu_list[0].data
        hdr = dict(hdu_list[0].header)
        hdu_list.close()

        # If we want to normalize the spectra, we must do it now since later we won't have the full EM range
        if norm:
            flux *= 1e-8  # convert from erg/cm^2/s/cm to erg/cm^2/s/A
            F_bol = np.trapz(flux, self.wl_full)
            # bolometric luminosity is always 1 L_sun
            flux *= C.F_sun / F_bol

        # Add temp, logg, Z, alpha, norm to the metadata
        hdr["norm"] = norm
        hdr["air"] = self.air

        if header:
            return (flux[self.ind], hdr)
        else:
            return flux[self.ind]


class PHOENIXGridInterfaceNoAlpha(PHOENIXGridInterface):
    """
    An Interface to the PHOENIX/Husser synthetic library without any Alpha concentration doping.

    Parameters
    ----------
    path : str or path-like
        The path of the base of the PHOENIX library

    Keyword Arguments
    -----------------
    kwargs : dict
        Any additional arguments will be passed to :class:`PHOENIXGridInterface`'s constructor

    See Also
    --------
    :class:`PHOENIXGridInterface`
    """

    def __init__(self, path, **kwargs):
        # Initialize according to the regular PHOENIX values
        super().__init__(path=path, **kwargs)

        # Now override parameters to exclude alpha
        self.param_names = ["T", "logg", "Z"]
        del self.points[3]
        del self.par_dicts[3]

        self.rname = (
            "Z{2:}/lte{0:0>5.0f}-{1:.2f}{2:}.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits"
        )
        self.full_rname = os.path.join(self.path, self.rname)

    def check_params(self, parameters):
        # Bypass the alpha irregularities in the full grid
        parameters = np.asarray(parameters)

        if len(parameters) != len(self.param_names):
            raise ValueError(
                "Length of given parameters ({}) does not match length of grid parameters ({})".format(
                    len(parameters), len(self.param_names)
                )
            )

        for param, params in zip(parameters, self.points):
            if param not in params:
                raise ValueError("{} not in the grid points {}".format(param, params))
        return True


class KuruczGridInterface(GridInterface):
    """
    Kurucz grid interface.

    Spectra are stored in ``f_nu`` in a filename like
    ``t03500g00m25ap00k2v070z1i00.fits``, ``ap00`` means zero alpha enhancement,
    and ``k2`` is the microturbulence, while ``z1`` is the macroturbulence.
    These particular values are roughly the ones appropriate for the Sun.
    """

    def __init__(self, path, air=True, wl_range=(5000, 5400)):
        super().__init__(
            name="Kurucz",
            param_names=["T", "logg", "Z"],
            points=[
                np.arange(3500, 9751, 250),
                np.arange(0.0, 5.1, 0.5),
                np.arange(-2.5, 0.51, 0.5),
            ],
            wave_units="AA",
            flux_units="",
            air=air,
            wl_range=wl_range,
            path=path,
        )

        self.par_dicts = [
            None,
            None,
            {
                -2.5: "m25",
                -2.0: "m20",
                -1.5: "m15",
                -1.0: "m10",
                -0.5: "m05",
                0.0: "p00",
                0.5: "p05",
            },
        ]

        # Convert to f_lam and average to 1, or leave in f_nu?
        self.rname = (
            "t{0:0>5.0f}/g{1:0>2.0f}/t{0:0>5.0f}g{1:0>2.0f}{2}ap00k2v000z1i00.fits"
        )
        self.full_rname = os.path.join(self.path, self.rname)
        self.wl_full = np.load(os.path.join(path, "kurucz_raw_wl.npy"))
        self.ind = (self.wl_full >= self.wl_range[0]) & (
            self.wl_full <= self.wl_range[1]
        )
        self.wl = self.wl_full[self.ind]

    def load_flux(self, parameters, header=False, norm=True):
        self.check_params(parameters)

        str_parameters = []
        for param, par_dict in zip(parameters, self.par_dicts):
            if par_dict is None:
                str_parameters.append(param)
            else:
                str_parameters.append(par_dict[param])

        # Multiply logg by 10
        str_parameters[1] *= 10

        fname = self.full_rname.format(*str_parameters)

        # Still need to check that file is in the grid, otherwise raise a C.GridError
        # Read all metadata in from the FITS header, and append to spectrum
        try:
            flux_file = fits.open(fname)
            f = flux_file[0].data
            hdr = dict(flux_file[0].header)
            flux_file.close()
        except:
            raise ValueError("{} is not on disk.".format(fname))

        # We cannot normalize the spectra, since we don't have a full wl range, so instead we set the average
        # flux to be 1

        # Also, we should convert from f_nu to f_lam
        if norm:
            f *= C.c_ang / self.wl**2  # Convert from f_nu to f_lambda
            f /= np.average(f)  # divide by the mean flux, so avg(f) = 1

        # Add temp, logg, Z, norm to the metadata
        hdr["norm"] = norm
        hdr["air"] = self.air

        if header:
            return (f[self.ind], hdr)
        else:
            return f[self.ind]

    @staticmethod
    def get_wl_kurucz(filename):
        """The Kurucz grid is log-linear spaced."""
        flux_file = fits.open(filename)
        hdr = flux_file[0].header
        num = len(flux_file[0].data)
        p = np.arange(num)
        w1 = hdr["CRVAL1"]
        dw = hdr["CDELT1"]
        wl = 10 ** (w1 + dw * p)
        return wl


class BTSettlGridInterface(GridInterface):
    """
    BTSettl grid interface. Unlike the PHOENIX and Kurucz grids, the
    individual files of the BTSettl grid do not always have the same wavelength
    sampling. Therefore, each call of :meth:`load_flux` will interpolate the
    flux onto a LogLambda spaced grid that ranges between `wl_range` and has a
    velocity spacing of 0.08 km/s or better.

    If you have a choice, it's probably easier to use the Husser PHOENIX grid.
    """

    def __init__(self, path, air=True, wl_range=(2999, 13000)):
        super().__init__(
            name="BTSettl",
            points={
                "T": np.arange(3000, 7001, 100),
                "logg": np.arange(2.5, 5.6, 0.5),
                "Z": np.arange(-0.5, 0.6, 0.5),
                "alpha": np.array([0.0]),
            },
            wave_units="AA",
            flux_units="",
            air=air,
            wl_range=wl_range,
            path=path,
        )

        # Normalize to 1 solar luminosity?
        self.rname = "CIFIST2011/M{Z:}/lte{T:0>3.0f}-{logg:.1f}{Z:}.BT-Settl.spec.7.bz2"
        self.full_rname = os.path.join(self.path, self.rname)
        # self.Z_dict = {-2:'-2.0', -1.5:'-1.5', -1:'-1.0', -0.5:'-0.5', 0.0: '-0.0', 0.5: '+0.5', 1: '+1.0'}
        self.Z_dict = {-0.5: "-0.5a+0.2", 0.0: "-0.0a+0.0", 0.5: "+0.5a0.0"}

        wl_dict = create_log_lam_grid(
            0.08 / C.c_kms, start=self.wl_range[0], end=self.wl_range[1]
        )
        self.wl = wl_dict["wl"]

    def load_flux(self, parameters, norm=True):
        """
        Because of the crazy format of the BTSettl, we need to sort the wl to make sure
        everything is unique, and we're not screwing ourselves with the spline.

        :param parameters: stellar parameters
        :type parameters: dict
        :param norm: If True, will normalize the spectrum to solar luminosity. Default is True
        :type norm: bool
        """
        # Check to make sure that the keys are allowed and that the values are in the grid
        self.check_params(parameters)

        str_parameters = parameters.copy()

        # Rewrite Z
        Z = parameters["Z"]
        str_parameters["Z"] = self.Z_dict[Z]

        # Multiply temp by 0.01
        str_parameters["T"] = 0.01 * parameters["T"]

        fname = self.full_rname.format(**str_parameters)
        file = bz2.BZ2File(fname, "r")

        lines = file.readlines()
        strlines = [line.decode("utf-8") for line in lines]
        file.close()

        data = ascii.read(
            strlines,
            col_starts=(0, 13),
            col_ends=(12, 25),
            Reader=ascii.FixedWidthNoHeader,
        )
        wl = data["col1"]
        fl_str = data["col2"]

        # convert because of 'D' exponent, unreadable in Python
        fl = idl_float(fl_str)
        fl = 10 ** (fl - 8.0)  # now in ergs/cm^2/s/A

        # 'Clean' the wl and flux points. Remove duplicates, sort in increasing wl
        wl, ind = np.unique(wl, return_index=True)
        fl = fl[ind]

        if norm:
            F_bol = np.trapz(fl, wl)
            fl = fl * (C.F_sun / F_bol)
            # the bolometric luminosity is always 1 L_sun

        # truncate the spectrum to the wl range of interest
        # at this step, make the range a little more so that the next stage of
        # spline interpolation is properly in bounds
        ind = (wl >= (self.wl_range[0] - 50.0)) & (wl <= (self.wl_range[1] + 50.0))
        wl = wl[ind]
        fl = fl[ind]

        if self.air:
            # Shift the wl that correspond to the raw spectrum
            wl = vacuum_to_air(wl)

        # Now interpolate wl, fl onto self.wl
        interp = InterpolatedUnivariateSpline(wl, fl, k=5)
        fl_interp = interp(self.wl)

        return fl_interp


class CIFISTGridInterface(GridInterface):
    """
    CIFIST grid interface, grid available here: https://phoenix.ens-lyon.fr/Grids/BT-Settl/CIFIST2011_2015/FITS/.
    Unlike the PHOENIX and Kurucz grids, the
    individual files of the BTSettl grid do not always have the same wavelength
    sampling. Therefore, each call of :meth:`load_flux` will interpolate the
    flux onto a LogLambda spaced grid that ranges between `wl_range` and has a
    velocity spacing of 0.08 km/s or better.

    If you have a choice, it's probably easier to use the Husser PHOENIX grid.
    """

    def __init__(self, path, air=True, wl_range=(3000, 13000)):
        super().__init__(
            name="CIFIST",
            points=[
                np.concatenate(
                    (np.arange(1200, 2351, 50), np.arange(2400, 7001, 100)), axis=0
                ),
                np.arange(2.5, 5.6, 0.5),
            ],
            param_names=["T", "logg"],
            wave_units="AA",
            flux_units="",
            air=air,
            wl_range=wl_range,
            path=path,
        )

        self.par_dicts = [None, None]
        self.rname = "lte{0:0>5.1f}-{1:.1f}-0.0a+0.0.BT-Settl.spec.fits.gz"
        self.full_rname = os.path.join(self.path, self.rname)

        wl_dict = create_log_lam_grid(
            dv=0.08, start=self.wl_range[0], end=self.wl_range[1]
        )
        self.wl = wl_dict["wl"]

    def load_flux(self, parameters, header=False, norm=True):
        self.check_params(parameters)

        str_parameters = []
        for param, par_dict in zip(parameters, self.par_dicts):
            if par_dict is None:
                str_parameters.append(param)
            else:
                str_parameters.append(par_dict[param])

        # Multiply temp by 0.01
        str_parameters[0] = 0.01 * parameters[0]

        fname = self.full_rname.format(*str_parameters)

        # Still need to check that file is in the grid, otherwise raise a C.GridError
        # Read all metadata in from the FITS header, and append to spectrum
        try:
            flux_file = fits.open(fname)
            data = flux_file[1].data
            hdr = dict(flux_file[1].header)

            wl = data["Wavelength"] * 1e4  # [Convert to angstroms]
            fl = data["Flux"]

            flux_file.close()
        except OSError:
            raise C.GridError("{} is not on disk.".format(fname))

        # "Clean" the wl and flux points. Remove duplicates, sort in increasing wl
        wl, ind = np.unique(wl, return_index=True)
        fl = fl[ind]

        if norm:
            F_bol = np.trapz(fl, wl)
            fl = fl * (C.F_sun / F_bol)
            # the bolometric luminosity is always 1 L_sun

        # truncate the spectrum to the wl range of interest
        # at this step, make the range a little more so that the next stage of
        # spline interpolation is properly in bounds
        ind = (wl >= (self.wl_range[0] - 50.0)) & (wl <= (self.wl_range[1] + 50.0))
        wl = wl[ind]
        fl = fl[ind]

        if self.air:
            # Shift the wl that correspond to the raw spectrum
            wl = vacuum_to_air(wl)

        # Now interpolate wl, fl onto self.wl
        interp = InterpolatedUnivariateSpline(wl, fl, k=5)
        fl_interp = interp(self.wl)

        # Add temp, logg, Z, norm to the metadata

        hdr["norm"] = norm
        hdr["air"] = self.air

        if header:
            return (fl_interp, hdr)
        else:
            return fl_interp


def load_BTSettl(T, logg, Z, norm=False, trunc=False, air=False):
    rname = "BT-Settl/CIFIST2011/M{Z:}/lte{T:0>3.0f}-{logg:.1f}{Z:}.BT-Settl.spec.7.bz2".format(
        T=0.01 * T, logg=logg, Z=Z
    )
    file = bz2.BZ2File(rname, "r")

    lines = file.readlines()
    strlines = [line.decode("utf-8") for line in lines]
    file.close()

    data = ascii.read(
        strlines, col_starts=[0, 13], col_ends=[12, 25], Reader=ascii.FixedWidthNoHeader
    )
    wl = data["col1"]
    fl_str = data["col2"]

    # convert because of "D" exponent, unreadable in Python
    fl = idl_float(fl_str)
    fl = 10 ** (fl - 8.0)  # now in ergs/cm^2/s/A

    if norm:
        F_bol = np.trapz(fl, wl)
        fl = fl * (C.F_sun / F_bol)
        # this also means that the bolometric luminosity is always 1 L_sun

    if trunc:
        # truncate to only the wl of interest
        ind = (wl > 3000) & (wl < 13000)
        wl = wl[ind]
        fl = fl[ind]

    if air:
        wl = vacuum_to_air(wl)

    return [wl, fl]



import itertools
import logging
import multiprocessing as mp
import os

import h5py
import numpy as np
from astropy.io import fits
import tqdm

import Starfish.constants as C
from Starfish.transforms import instrumental_broaden, resample
from Starfish.utils import calculate_dv, calculate_dv_dict, create_log_lam_grid
from .utils import chunk_list

log = logging.getLogger(__name__)


class GridInterface:
    """
    A base class to handle interfacing with synthetic spectral libraries.

    Parameters
    ----------
    path : str or path-like
        path to the root of the files on disk.
    param_names : list of str
        The names of the parameters (dimensions) of the grid
    points :  array_like
        the grid points at which
        spectra exist (assumes grid is square, not ragged, meaning that every combination
        of parameters specified exists in the grid).
    wave_units : str
        The units of the wavelengths. Preferably equivalent to an astropy unit string.
    flux_units : str
        The units of the model fluxes. Preferable equivalent to an astropy unit string.
    wl_range :  list [min, max], optional
        the starting and ending wavelength ranges of the grid to
        truncate to. If None, will use whole available grid. Default is None.
    air :  bool, optional
        Are the wavelengths measured in air? Default is True
    name : str, optional
        name of the spectral library, Default is None
    """

    def __init__(
        self,
        path,
        param_names,
        points,
        wave_units,
        flux_units,
        wl_range=None,
        air=True,
        name=None,
    ):
        self.path = path
        self.param_names = param_names
        self.points = points
        self.air = air
        self.wl_range = wl_range
        self.wave_units = wave_units
        self.flux_units = flux_units
        if name is None:
            name = "Grid Interface"
        self.name = name

    def check_params(self, parameters):
        """
        Determine if the specified parameters are allowed in the grid.

        Parameters
        ----------
        parameters : array_like
            parameter set to check

        Raises
        -------
        ValueError
            if the parameter values are outside of the grid bounds

        Returns
        -------
        bool
            True if found in grid
        """
        parameters = np.asarray(parameters)

        if len(parameters) != len(self.param_names):
            raise ValueError(
                "Length of given parameters ({}) does not match length of grid parameters ({})".format(
                    len(parameters), len(self.param_names)
                )
            )

        for param, params in zip(parameters, self.points):
            if param not in params:
                raise ValueError("{} not in the grid points {}".format(param, params))
        return True

    def load_flux(self, parameters, header=False, norm=True):
        """
        Load the flux and header information.

        Parameters
        ----------
        parameters : array_like
            stellar parameters
        header : bool, optional
            If True, will return the header alongside the flux. Default is False.
        norm : bool, optional
            If True, will normalize the flux to solar luminosity. Default is True.

        Raises
        ------
        ValueError
            if the file cannot be found on disk.

        Returns
        -------
        numpy.ndarray if header is False, tuple of (numpy.ndarray, dict) if header is True
        """
        raise NotImplementedError(
            "`load_flux` is abstract and must be implemented by subclasses"
        )

    def __repr__(self):
        output = "{}\n".format(self.name)
        output += "-" * len(self.name) + "\n"
        output += "Base: {}\n".format(self.path)
        for par, point in zip(self.param_names, self.points):
            output += "{}: {}\n".format(par, point)
        return output


class HDF5Interface:
    """
    Connect to an HDF5 file that stores spectra.

    Parameters
    ----------
    filename : str or path-like
        The path of the saved HDF5 file
    """

    def __init__(self, filename):
        self.filename = os.path.expandvars(filename)

        # In order to properly interface with the HDF5 file, we need to learn
        # a few things about it

        # 1.) Which parameter combinations exist in the file (self.grid_points)
        # 2.) What are the minimum and maximum values for each parameter (self.bounds)
        # 3.) Which values exist for each parameter (self.points)

        with h5py.File(self.filename, "r") as base:
            self.wl = base["wl"][:]
            self.key_name = base["flux"].attrs["key_name"]
            self.wl_header = dict(base["wl"].attrs.items())
            self.dv = self.wl_header["dv"]
            self.grid_points = base["grid_points"][:]
            self.param_names = base["grid_points"].attrs["names"]
            self.wave_units = base["wl"].attrs["units"]
            self.flux_units = base["flux"].attrs["units"]

        # determine the bounding regions of the grid by sorting the grid_points
        low = np.min(self.grid_points, axis=0)
        high = np.max(self.grid_points, axis=0)
        self.bounds = np.vstack((low, high)).T
        self.points = [
            np.unique(self.grid_points[:, i]) for i in range(self.grid_points.shape[1])
        ]

        self.ind = None  # Overwritten by other methods using this as part of a ModelInterpolator

        # Test if key-name is specified correctly
        try:
            self.load_flux(self.grid_points[0])
        except (IndexError, KeyError):
            raise ValueError("key_name is ill-specified.")

    def load_flux(self, parameters, header=False):
        """
        Load just the flux from the grid, with possibly an index truncation.

        parameters : array_like
            the stellar parameters
        header : bool, optional
            If True, will return the header as well as the flux. Default is False


        Returns
        -------
        numpy.ndarray if header is False, otherwise (numpy.ndarray, dict)
        """
        parameters = np.asarray(parameters)

        key = self.key_name.format(*parameters)
        with h5py.File(self.filename, "r") as hdf5:
            hdr = dict(hdf5["flux"][key].attrs)
            if self.ind is not None:
                fl = hdf5["flux"][key][self.ind[0] : self.ind[1]]
            else:
                fl = hdf5["flux"][key][:]

        # Note: will raise a KeyError if the file is not found.
        if header:
            return fl, hdr
        else:
            return fl

    @property
    def fluxes(self):
        """
        Iterator to loop over all of the spectra stored in the grid, for PCA.
        Loops over parameters in the order specified by grid_points.

        Returns
        -------
        Generator of numpy.ndarrays
        """

        for grid_point in self.grid_points:
            yield self.load_flux(grid_point, header=False)


class HDF5Creator:
    """
    Create a HDF5 grid to store all of the spectra from a RawGridInterface,
    along with metadata.

    Parameters
    ----------
    grid_interface : :class:`GridInterface`
        The raw grid interface to process while creating the HDF5 file
    filename : str or path-like
        Where to save the HDF5 file
    instrument : :class:`Instrument`, optional
        If provided, the instrument to convolve/truncate the grid. If None, will
        maintain the grid's original wavelengths and resolution. Default is None
    wl_range : list [min, max], optional
        The wavelength range to truncate the grid to. Will be truncated to match grid wavelengths and instrument wavelengths if over or under specified. If set to None, will not truncate grid. Default is NOne
    ranges : array_like, optional
        lower and upper limits for each stellar parameter,
        in order to truncate the number of spectra in the grid. If None, will not restrict the range of the parameters. Default is None.
    key_name : format str
        formatting string that has keys for each of the parameter names to translate into a hash-able string. If set to None, will create a name by taking each parameter name followed by value with underscores delimiting parameters. Default is None.

    Raises
    ------
    ValueError
        if the wl_range is ill-specified or if the parameter range are completely disjoint from the grid points.
    """

    def __init__(
        self,
        grid_interface,
        filename,
        instrument=None,
        wl_range=None,
        ranges=None,
        key_name=None,
    ):

        self.log = logging.getLogger(self.__class__.__name__)

        self.grid_interface = grid_interface
        self.filename = os.path.expandvars(filename)
        self.instrument = instrument

        # The flux formatting key will always have alpha in the name, regardless
        # of whether or not the library uses it as a parameter.
        if key_name is None:
            key_name = (
                self.grid_interface.rname.replace("/", "__")
                .replace(".fits", "")
                .replace(".FITS", "")
            )

        self.key_name = key_name

        if ranges is None:
            self.points = self.grid_interface.points
        else:
            # Take only those points of the GridInterface that fall within the ranges specified
            self.points = []
            # We know which subset we want, so use these.
            for i, (low, high) in enumerate(ranges):
                valid_points = self.grid_interface.points[i]
                ind = (valid_points >= low) & (valid_points <= high)
                self.points.append(valid_points[ind])
                # Note that at this point, this is just the grid points that fall within the rectangular
                # bounds set by ranges. If the raw library is actually irregular (e.g. CIFIST),
                # then self.points will contain points that don't actually exist in the raw library.

        # the raw wl from the spectral library
        self.wl_native = self.grid_interface.wl  # raw grid
        self.dv_native = calculate_dv(self.wl_native)

        self.hdf5 = h5py.File(self.filename, "w")
        self.hdf5.attrs["grid_name"] = grid_interface.name
        self.flux_group = self.hdf5.create_group("flux")
        self.flux_group.attrs["units"] = grid_interface.flux_units
        self.flux_group.attrs["key_name"] = self.key_name

        # We'll need a few wavelength grids
        # 1. The original synthetic grid: ``self.wl_native``
        # 2. A finely spaced log-lambda grid respecting the ``dv`` of
        #   ``self.wl_native``, onto which we can interpolate the flux values
        #   in preperation of the FFT: ``self.wl_loglam``
        # [ DO FFT ]
        # 3. A log-lambda spaced grid onto which we can downsample the result
        #   of the FFT, spaced with a ``dv`` such that we respect the remaining
        #   Fourier modes: ``self.wl_final``

        # There are three ranges to consider when wanting to make a grid:
        # 1. The full range of the synthetic library
        # 2. The full range of the instrument/dataset
        # 3. The range specified by the user in config.yaml
        # For speed reasons, we will always truncate to to wl_range. If either
        # the synthetic library or the instrument library is smaller than this range,
        # raise an error.
        if wl_range is None:
            wl_min, wl_max = 0, np.inf
        else:
            wl_min, wl_max = wl_range
        buffer = 50  # [AA]
        wl_min -= buffer
        wl_max += buffer

        # If the raw synthetic grid doesn't span the full range of the user
        # specified grid, raise an error.
        # Instead, let's choose the maximum limit of the synthetic grid?
        if self.instrument is not None:
            inst_min, inst_max = self.instrument.wl_range
        else:
            inst_min, inst_max = 0, np.inf
        imposed_min = np.max([self.wl_native[0], inst_min])
        imposed_max = np.min([self.wl_native[-1], inst_max])
        if wl_min < imposed_min:
            self.log.info(
                "Given minimum wavelength ({}) is less than instrument or grid minimum. Truncating to {}".format(
                    wl_min, imposed_min
                )
            )
            wl_min = imposed_min
        if wl_max > imposed_max:
            self.log.info(
                "Given maximum wavelength ({}) is greater than instrument or grid maximum. Truncating to {}".format(
                    wl_max, imposed_max
                )
            )
            wl_max = imposed_max

        if wl_max < wl_min:
            raise ValueError("Minimum wavelength must be less than maximum wavelength")

        # Calculate wl_loglam
        # use the dv that preserves the native quality of the raw PHOENIX grid
        wl_dict = create_log_lam_grid(self.dv_native, wl_min, wl_max)
        wl_loglam = wl_dict["wl"]
        dv_loglam = calculate_dv_dict(wl_dict)

        self.log.info(
            "FFT grid stretches from {} to {}".format(wl_loglam[0], wl_loglam[-1])
        )
        self.log.info("wl_loglam dv is {} km/s".format(dv_loglam))

        if self.instrument is None:
            mask = (self.wl_native > wl_min) & (self.wl_native < wl_max)
            self.wl_final = self.wl_native[mask]
            self.dv_final = self.dv_native

            def inst_broaden(w, f):
                return (w, f)

        else:

            def inst_broaden(w, f):
                return (w, instrumental_broaden(w, f, self.instrument.FWHM))

            # The final wavelength grid, onto which we will interpolate the
            # Fourier filtered wavelengths, is part of the instrument object
            dv_temp = self.instrument.FWHM / self.instrument.oversampling
            wl_dict = create_log_lam_grid(dv_temp, wl_min, wl_max)
            self.wl_final = wl_dict["wl"]
            self.dv_final = calculate_dv_dict(wl_dict)

        def resample_loglam(w, f):
            return (wl_loglam, resample(w, f, wl_loglam))

        def resample_final(w, f):
            return (self.wl_final, resample(w, f, self.wl_final))

        self.transform = lambda flux: resample_final(
            *inst_broaden(*resample_loglam(self.wl_native, flux))
        )

        # Create the wl dataset separately using float64 due to rounding errors w/ interpolation.
        wl_dset = self.hdf5.create_dataset("wl", data=self.wl_final, compression=9)
        wl_dset.attrs["air"] = self.grid_interface.air
        wl_dset.attrs["dv"] = self.dv_final
        wl_dset.attrs["units"] = self.grid_interface.wave_units

    def process_grid(self):
        """
        Run :meth:`process_flux` for all of the spectra within the `ranges`
        and store the processed spectra in the HDF5 file.
        """

        # points is now a list of numpy arrays of the values in the grid
        # Take all parameter permutations in self.points and create a list
        # param_list will be a list of numpy arrays, specifying the parameters
        param_list = []

        # use itertools.product to create permutations of all possible values
        for i in itertools.product(*self.points):
            param_list.append(np.array(i))

        all_params = np.array(param_list)

        invalid_params = []

        self.log.debug("Total of {} files to process.".format(len(param_list)))

        pbar = tqdm.tqdm(all_params)
        for i, param in enumerate(pbar):
            pbar.set_description("Processing {}".format(param))
            # Load and process the flux
            try:
                flux, header = self.grid_interface.load_flux(param, header=True)
            except ValueError:
                self.log.warning(
                    "Deleting {} from all params, does not exist.".format(param)
                )
                invalid_params.append(i)
                continue

            _, fl_final = self.transform(flux)

            flux = self.flux_group.create_dataset(
                self.key_name.format(*param), data=fl_final, compression=9
            )
            # Store header keywords as attributes in HDF5 file
            for key, value in header.items():
                if (
                    key != "" and key != "COMMENT" and value != ""
                ):  # check for empty FITS kws
                    flux.attrs[key] = value

        # Remove parameters that do no exist
        all_params = np.delete(all_params, invalid_params, axis=0)

        gp = self.hdf5.create_dataset("grid_points", data=all_params, compression=9)
        names = list(map(lambda s: s.encode("utf-8"), self.grid_interface.param_names))
        gp.attrs["names"] = names

        self.hdf5.close()



from collections import deque
from typing import Union, Sequence, Optional
import logging

from flatdict import FlatterDict
import numpy as np
from scipy.linalg import cho_factor, cho_solve
from scipy.optimize import minimize
import toml

from Starfish import Spectrum
from Starfish.emulator import Emulator
from Starfish.transforms import (
    chebyshev_correct,
    rotational_broaden,
    resample,
    doppler_shift,
    extinct,
    rescale,
    _get_renorm_factor,
)
from Starfish.utils import calculate_dv, create_log_lam_grid
from .kernels import global_covariance_matrix, local_covariance_matrix


class SpectrumModel:
    """
    A single-order spectrum model.

    Parameters
    ----------
    emulator : :class:`Starfish.emulators.Emulator`
        The emulator to use for this model.
    data : :class:`Starfish.spectrum.Spectrum`
        The data to use for this model
    grid_params : array-like
        The parameters that are used with the associated emulator
    max_deque_len : int, optional
        The maximum number of residuals to retain in a deque of residuals. Default is
        100
    norm : bool, optional
        If true, will rescale the model flux to the appropriate flux normalization
        according to the original spectral library. Default is `False`.
    name : str, optional
        A name for the model. Default is 'SpectrumModel'

    Keyword Arguments
    -----------------
    params : dict
        Any remaining keyword arguments will be interpreted as parameters.


    Here is a table describing the avialable parameters and their related functions

    =========== ===============================================
     Parameter                 Function
    =========== ===============================================
    vsini        :func:`~Starfish.transforms.rotational_broaden`
    vz           :func:`~Starfish.transforms.doppler_shift`
    Av           :func:`~Starfish.transforms.extinct`
    Rv           :func:`~Starfish.transforms.extinct`
    log_scale    :func:`~Starfish.transforms.rescale`
    cheb         :func:`~Starfish.transforms.chebyshev_correct`
    =========== ===============================================

    .. note::
        If :attr:`log_scale` is not specified, the model will use
        :func:`~Starfish.transforms.renorm` to automatically scale the spectrum to the
        data using the ratio of integrated fluxes.

    .. note::
        `cheb` corresponds to a list/array of coefficients, however we force the constant
        coefficient (`c0`) to be 1. This means `cheb` will correspond to `c1, c2, ...`.
        The entire list can be retrieved like `model["cheb"]` and indiviual values can be
        retrieved with `model["cheb:1"]`.

    The ``global_cov`` keyword arguments must be a dictionary definining the
    hyperparameters for the global covariance kernel,
    :meth:`kernels.global_covariance_matrix`

    ================ =============================================================
    Global Parameter  Description
    ================ =============================================================
    log_amp          The natural logarithm of the amplitude of the Matern kernel
    log_ls           The natural logarithm of the lengthscale of the Matern kernel
    ================ =============================================================

    The ``local_cov`` keryword argument must be a list of dictionaries defining
    hyperparameters for many Gaussian kernels, , :meth:`kernels.local_covariance_matrix`

    ================ =============================================================
    Local Parameter  Description
    ================ =============================================================
    log_amp          The natural logarithm of the amplitude of the kernel
    mu               The location of the local kernel
    log_sigma        The natural logarithm of the standard deviation of the kernel
    ================ =============================================================

    Attributes
    ----------
    params : dict
        The dictionary of parameters that are used for doing the modeling. (The Chebyshev coefficients are not stored in this structure)
    grid_params : ndarray
        The vector of parameters for the spectral emulator. Setter obeys frozen parameters.
    cheb : ndarray
        The vector of `c1, c2, ...` Chebyshev coefficients. `c0` is fixed to 1 by definition. Setter obeys frozen parameters.
    frozen : list
        A list of strings corresponding to frozen parameters
    residuals : deque
        A deque containing residuals from calling :meth:`SpectrumModel.log_likelihood`
    """

    _PARAMS = [
        "vz",
        "vsini",
        "Av",
        "Rv",
        "log_scale",
        "global_cov",
        "local_cov",
        "cheb",
    ]
    _GLOBAL_PARAMS = ["log_amp", "log_ls"]
    _LOCAL_PARAMS = ["mu", "log_amp", "log_sigma"]

    def __init__(
        self,
        emulator: Union[str, Emulator],
        data: Union[str, Spectrum],
        grid_params: Sequence[float],
        max_deque_len: int = 100,
        norm=False,
        name: str = "SpectrumModel",
        **params,
    ):
        if isinstance(emulator, str):
            emulator = Emulator.load(emulator)
        if isinstance(data, str):
            data = Spectrum.load(data)

        if len(data) > 1:
            raise ValueError(
                "Multiple orders detected in data, please use EchelleModel"
            )

        self.emulator: Emulator = emulator
        self.data_name = data.name
        self.data = data[0]

        dv = calculate_dv(self.data.wave)
        self.min_dv_wave = create_log_lam_grid(
            dv, self.emulator.wl.min(), self.emulator.wl.max()
        )["wl"]
        self.bulk_fluxes = resample(
            self.emulator.wl, self.emulator.bulk_fluxes, self.min_dv_wave
        )

        self.residuals = deque(maxlen=max_deque_len)

        # manually handle cheb coeffs to offset index by 1
        if "cheb" in params:
            chebs = params.pop("cheb")
            cheb_idxs = [str(i) for i in range(1, len(chebs) + 1)]
            params["cheb"] = dict(zip(cheb_idxs, chebs))
        # load rest of params into FlatterDict
        self.params = FlatterDict(params)
        self.frozen = []
        self.name = name
        self.norm = norm

        # Unpack the grid parameters
        self.n_grid_params = len(grid_params)
        self.grid_params = grid_params

        # None means "yet to be calculated", do not use NaN
        self._lnprob = None
        self._glob_cov = None
        self._loc_cov = None
        self._log_scale = params.get("log_scale", None)

        self.log = logging.getLogger(self.__class__.__name__)

    @property
    def grid_params(self):
        """
        numpy.ndarray : The parameters used for the spectral emulator.

        :setter: Sets the values in the order of ``Emulator.param_names``
        """
        values = []
        for key in self.emulator.param_names:
            values.append(self.params[key])
        return np.array(values)

    @grid_params.setter
    def grid_params(self, values):
        for key, value in zip(self.emulator.param_names, values):
            if key not in self.frozen:
                self.params[key] = value

    @property
    def cheb(self):
        """
        numpy.ndarray : The Chebyshev polynomial coefficients used for the background model
        """
        return np.array(self.params["cheb"].values())

    @cheb.setter
    def cheb(self, values):
        if "cheb" in self.frozen:
            return
        for key, value in zip(self.params["cheb"], values):
            if key not in self.frozen:
                self.params["cheb"][key] = value

    @property
    def labels(self):
        """
        tuple of str : The thawed parameter names
        """
        keys = self.get_param_dict(flat=True).keys()
        return tuple(keys)

    def __getitem__(self, key):
        if key == "cheb":
            return list(self.params[key].values())
        else:
            return self.params[key]

    def __setitem__(self, key, value):
        if ":" in key:
            group, rest = key.split(":", 1)
            k = rest.split(":")[-1] if ":" in rest else rest
            if group == "global_cov" and k in self._GLOBAL_PARAMS:
                self.params[key] = value
            elif group == "local_cov" and k in self._LOCAL_PARAMS:
                self.params[key] = value
            elif group == "cheb":
                # widen list with 0s if necessary
                if "cheb" in self.params:
                    N = len(self.params[group])
                    idx = int(rest)
                    if idx == 0:
                        raise KeyError("cannot change constant Chebyshev term")
                    for i in range(N + 1, idx + 1):
                        self.params[f"{group}:{i}"] = 0
                self.params[key] = value
            else:
                raise KeyError(f"{key} not recognized")
        else:
            if key == "cheb":
                cheb_idxs = [str(i) for i in range(1, len(value) + 1)]
                self.params[key] = dict(zip(cheb_idxs, value))
            elif key in [*self._PARAMS, *self.emulator.param_names]:
                self.params[key] = value
            else:
                raise KeyError(f"{key} not recognized")

    def __delitem__(self, key):
        if key not in self.params:
            raise KeyError(f"{key} not in params")
        elif key == "global_cov":
            self._glob_cov = None
            self.frozen = [
                key for key in self.frozen if not key.startswith("global_cov")
            ]
        elif key == "local_cov":
            self._loc_cov = None
            self.frozen = [
                key for key in self.frozen if not key.startswith("local_cov")
            ]

        del self.params[key]
        if key in self.frozen:
            self.frozen.remove(key)

    def __call__(self):
        """
        Performs the transformations according to the parameters available in
        ``self.params``

        Returns
        -------
        flux, cov : tuple
            The transformed flux and covariance matrix from the model
        """
        wave = self.min_dv_wave
        fluxes = self.bulk_fluxes

        if "vsini" in self.params:
            fluxes = rotational_broaden(wave, fluxes, self.params["vsini"])

        if "vz" in self.params:
            wave = doppler_shift(wave, self.params["vz"])

        fluxes = resample(wave, fluxes, self.data.wave)

        if "Av" in self.params:
            fluxes = extinct(self.data.wave, fluxes, self.params["Av"])

        if "cheb" in self.params:
            # force constant term to be 1 to avoid degeneracy with log_scale
            coeffs = [1, *self.cheb]
            fluxes = chebyshev_correct(self.data.wave, fluxes, coeffs)

        weights, weights_cov = self.emulator(self.grid_params)

        # Decompose the bulk_fluxes (see emulator/emulator.py for the ordering)
        *eigenspectra, flux_mean, flux_std = fluxes

        # Complete the reconstruction
        X = eigenspectra * flux_std
        flux = weights @ X + flux_mean

        # optionally scale using absolute flux calibration
        if self.norm:
            norm = self.emulator.norm_factor(self.grid_params)
        else:
            norm = 1

        # Renorm to data flux if no "log_scale" provided
        if "log_scale" not in self.params:
            scale = _get_renorm_factor(self.data.wave, flux * norm, self.data.flux)
            self._log_scale = np.log(scale)
            scale *= norm
            self.log.debug(f"fit scale factor using integrated flux ratio: {scale}")
        else:
            self._log_scale = self.params["log_scale"]
            scale = np.exp(self.params["log_scale"]) * norm

        flux = rescale(flux, scale)
        X = rescale(X, scale)

        L, flag = cho_factor(weights_cov, overwrite_a=True)
        cov = X.T @ cho_solve((L, flag), X)

        # Trivial covariance
        np.fill_diagonal(cov, cov.diagonal() + self.data.sigma**2)

        # Global covariance
        if "global_cov" in self.params:
            if "global_cov" not in self.frozen or self._glob_cov is None:
                ag = np.exp(self.params["global_cov:log_amp"])
                lg = np.exp(self.params["global_cov:log_ls"])
                self._glob_cov = global_covariance_matrix(self.data.wave, ag, lg)

        if self._glob_cov is not None:
            cov += self._glob_cov

        # Local covariance
        if "local_cov" in self.params:
            if "local_cov" not in self.frozen or self._loc_cov is None:
                self._loc_cov = 0
                for kernel in self.params.as_dict()["local_cov"]:
                    mu = kernel["mu"]
                    amplitude = np.exp(kernel["log_amp"])
                    sigma = np.exp(kernel["log_sigma"])
                    self._loc_cov += local_covariance_matrix(
                        self.data.wave, amplitude, mu, sigma
                    )

        if self._loc_cov is not None:
            cov += self._loc_cov

        return flux, cov

    def log_likelihood(self, priors: Optional[dict] = None) -> float:
        """
        Returns the log probability of a multivariate normal distribution

        Parameters
        ----------
        priors : dict, optional
            If provided, will use these priors in the MLE. Should contain keys that
            match the model's keys and values that have a `logpdf` method that takes
            one value (like ``scipy.stats`` distributions). Default is None.

        Warning
        -------
        No checks will be done on the :attr:`priors` for speed.

        Returns
        -------
        float
        """
        # Priors
        prior_lp = 0

        if priors is not None:
            for key, prior in priors.items():
                if key in self.params:
                    prior_lp += prior.logpdf(self[key])

        if not np.isfinite(prior_lp):
            return -np.inf

        # Likelihood
        flux, cov = self()
        np.fill_diagonal(cov, cov.diagonal() + 1e-10)
        factor, flag = cho_factor(cov, overwrite_a=True)
        logdet = 2 * np.sum(np.log(factor.diagonal()))
        R = flux - self.data.flux
        self.residuals.append(R)
        sqmah = R @ cho_solve((factor, flag), R)
        self._lnprob = -(logdet + sqmah) / 2

        return self._lnprob + prior_lp

    def get_param_dict(self, flat: bool = False) -> dict:
        """
        Gets the dictionary of thawed parameters.

        Parameters
        ----------
        flat : bool, optional
            If True, returns the parameters completely flat. For example,
            ``['local']['0']['mu']`` would have the key ``'local:0:mu'``.
            Default is False

        Returns
        -------
        dict

        See Also
        --------
        :meth:`set_param_dict`
        """
        params = FlatterDict()
        for key, val in self.params.items():
            if key not in self.frozen:
                params[key] = val

        output = params if flat else params.as_dict()

        return output

    def set_param_dict(self, params):
        """
        Sets the parameters with a dictionary. Note that this should not be used to add
        new parameters

        Parameters
        ----------
        params : dict
            The new parameters. If a key is present in ``self.frozen`` it will not be
            changed

        See Also
        --------
        :meth:`get_param_dict`
        """
        params = FlatterDict(params)
        for key, val in params.items():
            if key not in self.frozen:
                self.params[key] = val

    def get_param_vector(self):
        """
        Get a numpy array of the thawed parameters

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        :meth:`set_param_vector`
        """
        return np.array(list(self.get_param_dict(flat=True).values()))

    def set_param_vector(self, params):
        """
        Sets the parameters based on the current thawed state. The values will be
        inserted according to the order of :obj:`SpectrumModel.labels`.

        Parameters
        ----------
        params : array_like
            The parameters to set in the model

        Raises
        ------
        ValueError
            If the `params` do not match the length of the current thawed parameters.

        See Also
        --------
        :meth:`get_param_vector`
        """
        if len(params) != len(self.labels):
            raise ValueError("Param Vector does not match length of thawed parameters")
        param_dict = dict(zip(self.labels, params))
        self.set_param_dict(param_dict)

    def freeze(self, names):
        """
        Freeze the given parameter such that :meth:`get_param_dict` and
        :meth:`get_param_vector` no longer include this parameter, however it will
        still be used when calling the model.

        Parameters
        ----------
        name : str or array-like
            The parameter to freeze. If ``'all'``, will freeze all parameters. If
            ``'global_cov'`` will freeze all global covariance parameters. If
            ``'local_cov'`` will freeze all local covariance parameters.

        Raises
        ------
        ValueError
            If the given parameter does not exist

        See Also
        --------
        :meth:`thaw`
        """
        names = np.atleast_1d(names)
        if names[0] == "all":
            for key in self.labels:
                if key not in self.frozen:
                    self.frozen.append(key)
            if "global_cov" in self.params:
                self.frozen.append("global_cov")
            if "local_cov" in self.params:
                self.frozen.append("local_cov")
            if "cheb" in self.params:
                self.frozen.append("cheb")
        else:
            for _name in names:
                # Avoid kookyness of numpy.str type
                name = str(_name)
                if name == "global_cov" or name == "cheb":
                    self.frozen.append(name)
                    if name == "global_cov":
                        self._glob_cov = None
                    for key in self.params.as_dict()[name].keys():
                        flat_key = f"{name}:{key}"
                        if flat_key not in self.frozen:
                            self.frozen.append(flat_key)
                elif name == "local_cov":
                    self.frozen.append("local_cov")
                    self._loc_cov = None
                    for i, kern in enumerate(self.params.as_dict()["local_cov"]):
                        for key in kern.keys():
                            flat_key = f"local_cov:{i}:{key}"
                            if flat_key not in self.frozen:
                                self.frozen.append(flat_key)
                elif name not in self.frozen and name in self.params:
                    self.frozen.append(name)

    def thaw(self, names):
        """
        Thaws the given parameter. Opposite of freezing

        Parameters
        ----------
        name : str or array-like
            The parameter to thaw. If ``'all'``, will thaw all parameters. If
            ``'global_cov'`` will thaw all global covariance parameters. If
            ``'local_cov'`` will thaw all local covariance parameters.

        Raises
        ------
        ValueError
            If the given parameter does not exist.

        See Also
        --------
        :meth:`freeze`
        """
        names = np.atleast_1d(names)
        if names[0] == "all":
            self.frozen = []
        else:
            for _name in names:
                # Avoid kookyness of numpy.str type
                name = str(_name)
                if name == "global_cov" or name == "cheb":
                    self.frozen.remove(name)
                    for key in self.params.as_dict()[name].keys():
                        flat_key = f"{name}:{key}"
                        self.frozen.remove(flat_key)
                elif name == "local_cov":
                    self.frozen.remove("local_cov")
                    for i, kern in enumerate(self.params.as_dict()["local_cov"]):
                        for key in kern.keys():
                            flat_key = f"local_cov:{i}:{key}"
                            self.frozen.remove(flat_key)
                elif name in self.frozen:
                    self.frozen.remove(name)

    def save(self, filename, metadata=None):
        """
        Saves the model as a set of parameters into a TOML file

        Parameters
        ----------
        filename : str or path-like
            The TOML filename to save to.
        metadata : dict, optional
            If provided, will save the provided dictionary under a 'metadata' key. This
            will not be read in when loading models but provides a way of providing
            information in the actual TOML files. Default is None.
        """
        output = {"parameters": self.params.as_dict(), "frozen": self.frozen}
        meta = {}
        meta["name"] = self.name
        meta["data"] = self.data_name
        if self.emulator.name is not None:
            meta["emulator"] = self.emulator.name
        if metadata is not None:
            meta.update(metadata)
        output["metadata"] = meta

        with open(filename, "w") as handler:
            encoder = toml.TomlNumpyEncoder(output.__class__)
            toml.dump(output, handler, encoder=encoder)

        self.log.info(f"Saved current state at {filename}")

    def load(self, filename):
        """
        Load a saved model state from a TOML file

        Parameters
        ----------
        filename : str or path-like
            The saved state to load
        """
        with open(filename, "r") as handler:
            data = toml.load(handler)
        self.params = FlatterDict(data["parameters"])
        self.frozen = data["frozen"]

    def train(self, priors: Optional[dict] = None, **kwargs):
        """
        Given a :class:`SpectrumModel` and a dictionary of priors, will perform
        maximum-likelihood estimation (MLE). This will use ``scipy.optimize.minimize`` to
        find the maximum a-posteriori (MAP) estimate of the current model state. Note
        that this alters the state of the model. This means that you can run this
        method multiple times until the optimization succeeds. By default, we use the
        "Nelder-Mead" method in `minimize` to avoid approximating any derivatives.

        Parameters
        ----------
        priors : dict, optional
            Priors to pass to :meth:`log_likelihood`
        **kwargs : dict, optional
            These keyword arguments will be passed to `scipy.optimize.minimize`

        Returns
        -------
        soln : `scipy.optimize.minimize_result`
            The output of the minimization.

        Raises
        ------
        ValueError
            If the priors are poorly specified
        RuntimeError
            If any priors evaluate to non-finite values

        See Also
        --------
        :meth:`log_likelihood`
        """
        if priors is None:
            priors = {}

        # Check priors for validity
        for key, val in priors.items():
            # Key exists
            if key not in self.params and not key.startswith("cheb"):
                raise ValueError(f"Invalid priors. {key} not a vlid key.")
            # has logpdf method
            if not callable(getattr(val, "logpdf", None)):
                raise ValueError(
                    f"Invalid priors. {key} does not have a `logpdf` method"
                )
            # Evaluates to a finite number in current state
            log_prob = val.logpdf(self[key])
            if not np.isfinite(log_prob):
                raise RuntimeError(f"{key}'s logpdf evaluated to {log_prob}")

        def nll(P):
            self.set_param_vector(P)
            return -self.log_likelihood(priors)

        p0 = self.get_param_vector()
        params = {"method": "Nelder-Mead"}
        params.update(kwargs)
        soln = minimize(nll, p0, **params)
        if soln.success:
            self.set_param_vector(soln.x)

        return soln

    def plot(self, axes=None, plot_kwargs=None, resid_kwargs=None):
        """
        Plot the model.

        This will create two subplots, one which shows the current model against the
        data, and another which shows the current residuals with 3:math:`\\sigma`
        contours from the diagonal of the covariance matrix. Note this requires
        matplotlib to be installed, which is not installed by default with Starfish.

        Parameters
        ----------
        axes : iterable of matplotlib.Axes, optional
            If provided, will use the first two axes to plot, otherwise will create new
            axes, by default None
        plot_kwargs : dict, optional
            If provided, will use these kwargs for the comparison plot, by default None
        resid_kwargs : dict, optional
            If provided, will use these kwargs for the residuals plot, by default None

        Returns
        -------
        list of matplotlib.Axes
            The returned axes, for the user to edit as they please
        """
        import matplotlib.pyplot as plt
        from matplotlib import rcParams

        if axes is None:
            # Set up a 4x4 grid with the main plot taking the whole left column
            figsize = rcParams["figure.figsize"]
            plt.figure(figsize=(figsize[0] * 1.75, figsize[1] * 1.1))
            grid = plt.GridSpec(2, 2, width_ratios=(1.25, 1))
            axes = [
                plt.subplot(grid[:, 0]),
                plt.subplot(grid[0, 1]),
                plt.subplot(grid[1, 1]),
            ]
            axes[1].tick_params(labelbottom=False)
        if plot_kwargs is None:
            plot_kwargs = {}
        if resid_kwargs is None:
            resid_kwargs = {}

        model_flux, model_cov = self()

        # Comparison plot
        plot_params = {"lw": 0.7}
        plot_params.update(plot_kwargs)
        ax = axes[0]
        ax.plot(self.data.wave, self.data.flux, label="Data", **plot_params)
        ax.plot(self.data.wave, model_flux, label="Model", **plot_params)
        ax.set_yscale("log")
        ax.set_xlabel(r"$\lambda$ [$\AA$]")
        ax.set_ylabel(r"$f_\lambda$ [$erg/cm^2/s/cm$]")
        ax.legend()

        # Residuals plot
        R = self.data.flux - model_flux
        std = np.sqrt(model_cov.diagonal())
        resid_params = {"lw": 0.3}
        resid_params.update(resid_kwargs)
        ax = axes[1]
        ax.plot(self.data.wave, R, c="k", label="Data - Model", **resid_params)
        ax.fill_between(
            self.data.wave, -std, std, color="C2", alpha=0.6, label=r"$\sigma$"
        )
        ax.fill_between(
            self.data.wave, -2 * std, 2 * std, color="C2", alpha=0.4, label=r"$2\sigma$"
        )
        ax.fill_between(
            self.data.wave, -3 * std, 3 * std, color="C2", alpha=0.2, label=r"$3\sigma$"
        )
        ax.set_ylabel(r"$\Delta f_\lambda$")
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position("right")
        ax.legend()

        # Relative Error plot
        R_f = R / self.data.flux
        ax = axes[2]
        ax.plot(self.data.wave, R_f, label="Data - Model", c="k", **resid_params)
        ax.set_xlabel(r"$\lambda$ [$\AA$]")
        ax.set_ylabel(r"$\Delta f_\lambda / f_\lambda$")
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position("right")

        plt.suptitle(self.data_name)
        plt.tight_layout(rect=[0, 0.03, 1, 0.95])
        return axes

    def __repr__(self):
        output = f"{self.name}\n"
        output += "-" * len(self.name) + "\n"
        output += f"Data: {self.data_name}\n"
        output += f"Emulator: {self.emulator.name}\n"
        output += f"Log Likelihood: {self._lnprob}\n"
        output += "\nParameters\n"
        for key, value in self.get_param_dict().items():
            if key == "global_cov":
                output += "  global_cov:\n"
                for gkey, gval in value.items():
                    output += f"    {gkey}: {gval}\n"
            elif key == "local_cov":
                output += "  local_cov:\n"
                for i, kern in enumerate(value.values()):
                    output += f"    {i}: "
                    for lkey, lval in kern.items():
                        output += f"{lkey}: {lval}, "
                    # Remove trailing whitespace and comma
                    output = output[:-2]
                    output += "\n"
            elif key == "cheb":
                output += f"  cheb: {list(value.values())}\n"
            else:
                output += f"  {key}: {value}\n"
        if "log_scale" not in self.params:
            self()
            output += f"  log_scale: {self._log_scale} (fit)\n"
        if len(self.frozen) > 0:
            output += "\nFrozen Parameters\n"
            for key in self.frozen:
                if key in ["global_cov", "local_cov"]:
                    continue
                output += f"  {key}: {self[key]}\n"
        return output[:-1]  # No trailing newline



import extinction  # This may be marked as unused, but is necessary
import numpy as np
from numpy.polynomial.chebyshev import chebval
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.special import j1

from Starfish.constants import c_kms
from Starfish.utils import calculate_dv


def resample(wave, flux, new_wave):
    """
    Resample onto a new wavelength grid using k=5 spline interpolation

    Parameters
    ----------
    wave : array_like
        The original wavelength grid
    flux : array_like
        The fluxes to resample
    new_wave : array_like
        The new wavelength grid

    Raises
    ------
    ValueError
        If the new wavelength grid is not strictly increasing monotonic

    Returns
    -------
    numpy.ndarray
        The resampled flux with the same 1st dimension as the input flux
    """

    if np.any(new_wave <= 0):
        raise ValueError("Wavelengths must be positive")

    if flux.ndim > 1:
        interpolators = [InterpolatedUnivariateSpline(wave, fl, k=5) for fl in flux]
        return np.array([interpolator(new_wave) for interpolator in interpolators])
    else:
        return InterpolatedUnivariateSpline(wave, flux, k=5)(new_wave)


def instrumental_broaden(wave, flux, fwhm):
    """
    Broadens given flux by convolving with a Gaussian kernel appropriate for a
    spectrograph's instrumental properties. Follows the given equation

    .. math::
        f = f * \\mathcal{F}^{\\text{inst}}_v

    .. math::
        \\mathcal{F}^{\\text{inst}}_v = \\frac{1}{\\sqrt{2\\pi \\sigma^2}} \\exp \\left[-\\frac12 \\left( \\frac{v}{\\sigma} \\right)^2 \\right]

    This is carried out by multiplication in the Fourier domain rather than using a
    convolution function.

    Parameters
    ----------
    wave : array_like
        The current wavelength grid
    flux : array_like
        The current flux
    fwhm : float
        The full width half-maximum of the instrument in km/s. Note that this is
        quivalent to :math:`2.355\\cdot \\sigma`

    Raises
    ------
    ValueError
        If the full width half maximum is negative.

    Returns
    -------
    numpy.ndarray
        The broadened flux with the same shape as the input flux
    """

    if fwhm < 0:
        raise ValueError("FWHM must be non-negative")
    dv = calculate_dv(wave)
    freq = np.fft.rfftfreq(flux.shape[-1], d=dv)
    flux_ff = np.fft.rfft(flux)

    sigma = fwhm / 2.355
    flux_ff *= np.exp(-2 * (np.pi * sigma * freq) ** 2)

    flux_final = np.fft.irfft(flux_ff, n=flux.shape[-1])
    return flux_final


def rotational_broaden(wave, flux, vsini):
    """
    Broadens flux according to a rotational broadening kernel from Gray (2005) [1]_

    Parameters
    ----------
    wave : array_like
        The current wavelength grid
    flux : array_like
        The current flux
    vsini : float
        The rotational velocity in km/s

    Raises
    ------
    ValueError
        if `vsini` is not positive

    Returns
    -------
    numpy.ndarray
        The broadened flux with the same shape as the input flux


    .. [1] Gray, D. (2005). *The observation and Analysis of Stellar Photospheres*.
    Cambridge: Cambridge University Press. doi:10.1017/CB09781316036570
    """

    if vsini <= 0:
        raise ValueError("vsini must be positive")

    dv = calculate_dv(wave)
    freq = np.fft.rfftfreq(flux.shape[-1], dv)
    flux_ff = np.fft.rfft(flux)
    # Calculate the stellar broadening kernel (Gray 2008)
    ub = 2.0 * np.pi * vsini * freq
    # Remove 0th frequency
    ub = ub[1:]
    sb = j1(ub) / ub - 3 * np.cos(ub) / (2 * ub**2) + 3.0 * np.sin(ub) / (2 * ub**3)
    flux_ff *= np.insert(sb, 0, 1.0)
    flux_final = np.fft.irfft(flux_ff, n=flux.shape[-1])
    return flux_final


def doppler_shift(wave, vz):
    """
    Doppler shift a spectrum according to the formula

    .. math::
        \\lambda \\cdot \\sqrt{\\frac{c + v_z}{c - v_z}}

    Parameters
    ----------
    wave : array_like
        The unshifted wavelengths
    vz : float
        The doppler velocity in km/s

    Returns
    -------
    numpy.ndarray
        Altered wavelengths with the same shape as the input wavelengths
    """

    dv = np.sqrt((c_kms + vz) / (c_kms - vz))
    return wave * dv


def extinct(wave, flux, Av, Rv=3.1, law="ccm89"):
    """
    Extinct a spectrum following one of many empirical extinction laws. This makes use
    of the `extinction` package. In general, it follows the form

    .. math:: f \\cdot 10^{-0.4 A_V \\cdot A_\\lambda(R_V)}

    Parameters
    ----------
    wave : array_like
        The input wavelengths in Angstrom
    flux : array_like
        The input fluxes
    Av : float
        The absolute attenuation
    Rv : float, optional
        The relative attenuation (the default is 3.1, which is the Milky Way average)
    law : str, optional
        The extinction law to use. One of `{'ccm89', 'odonnell94', 'calzetti00',
        'fitzpatrick99', 'fm07'}` (the default is 'ccm89')

    Raises
    ------
    ValueError
        If `law` does not match one of the availabe laws
    ValueError
        If Rv is not positive

    Returns
    -------
    numpy.ndarray
        The extincted fluxes, with same shape as input fluxes.
    """

    if law not in ["ccm89", "odonnell94", "calzetti00", "fitzpatrick99", "fm07"]:
        raise ValueError("Invalid extinction law given")
    if Rv <= 0:
        raise ValueError("Rv must be positive")

    law_fn = eval("extinction.{}".format(law))
    if law == "fm07":
        A_l = law_fn(wave.astype(np.double), Av)
    else:
        A_l = law_fn(wave.astype(np.double), Av, Rv)
    flux_final = flux * 10 ** (-0.4 * A_l)
    return flux_final


def rescale(flux, scale):
    """
    Rescale the given flux via the following equation

    .. math:: f \\cdot \\Omega

    Parameters
    ----------
    flux : array_like
        The input fluxes
    scale : float or array_like
        The scaling factor. If an array, must have same shape as the batch dimension of
        :attr:`flux`

    Returns
    -------
    numpy.ndarray
        The rescaled fluxes with the same shape as the input fluxes
    """
    scale = np.atleast_1d(scale)
    if len(scale) > 1:
        scale = scale[:, np.newaxis]
    return flux * scale


def renorm(wave, flux, reference_flux):
    """
    Renormalize one spectrum to another

    This uses the :meth:`rescale` function with a :attr:`log_scale` of

    .. math::

        \\log \\Omega = \\left. \\int{f^{*}(w) dw} \\middle/ \\int{f(w) dw} \\right.

    where :math:`f^{*}` is the reference flux, :math:`f` is the source flux, and the
    integrals are over a common wavelength grid

    Parameters
    ----------
    wave : array_like
        The wavelength grid for the source flux
    flux : array_like
        The flux for the source
    reference_flux : array_like
        The reference source to renormalize to

    Returns
    -------
    numpy.ndarray
        The renormalized flux
    """
    factor = _get_renorm_factor(wave, flux, reference_flux)
    return rescale(flux, factor)


def _get_renorm_factor(wave, flux, reference_flux):
    ref_int = np.trapz(reference_flux, wave)
    flux_int = np.trapz(flux, wave, axis=-1)
    return ref_int / flux_int


def chebyshev_correct(wave, flux, coeffs):
    """
    Multiply the input flux by a Chebyshev series in order to correct for
    calibration-level discrepancies.

    Parameters
    ----------
    wave : array-lioke
        Input wavelengths
    flux : array-like
        Input flux
    coeffs : array-like
        The coefficients for the chebyshev series.

    Returns
    -------
    numpy.ndarray
        The corrected flux

    Raises
    ------
    ValueError
        If only processing a single spectrum and the linear coefficient is not 1.
    """
    # have to scale wave to fit on domain [0, 1]
    coeffs = np.asarray(coeffs)
    if coeffs.ndim == 1 and coeffs[0] != 1:
        raise ValueError(
            "For single spectrum the linear Chebyshev coefficient (c[0]) must be 1"
        )

    scale_wave = wave / wave.max()
    p = chebval(scale_wave, coeffs, tensor=False)
    return flux * p



import numpy as np
from typing import List

from Starfish import constants as C


def global_covariance_matrix(
    wave: np.ndarray, amplitude: float, lengthscale: float
) -> np.ndarray:
    """
    A matern-3/2 kernel where the metric is defined as the velocity separation of the wavelengths.

    Parameters
    ----------
    wave : numpy.ndarray
        The wavelength grid
    amplitude : float
        The amplitude of the kernel
    lengthscale : float
        The lengthscale of the kernel

    Returns
    -------
    cov : numpy.ndarray
        The global covariance matrix
    """
    wx, wy = np.meshgrid(wave, wave)
    r = C.c_kms / 2 * np.abs((wx - wy) / (wx + wy))
    r0 = 6 * lengthscale

    # Calculate the kernel, being careful to stay in mask
    kernel = np.zeros((len(wx), len(wy)))
    mask = r <= r0
    taper = 0.5 + 0.5 * np.cos(np.pi * r[mask] / r0)
    kernel[mask] = (
        taper
        * amplitude
        * (1 + np.sqrt(3) * r[mask] / lengthscale)
        * np.exp(-np.sqrt(3) * r[mask] / lengthscale)
    )
    return kernel


def local_covariance_matrix(
    wave: np.ndarray, amplitude: float, mu: float, sigma: float
) -> np.ndarray:
    """
    A local Gaussian kernel. In general, the kernel has density like

    .. math::
        K(\\lambda | A, \\mu, \\sigma) = A \\exp\\left[-\\frac12 \\frac{\\left(\\lambda - \\mu\\right)^2}{\\sigma^2} \\right]

    Parameters
    ----------
    wave : numpy.ndarray
        The wavelength grid
    amplitude : float
        The amplitudes of the Gaussian
    mu : float
        The means of the Gaussian
    sigma : float
        The standard deviations of the Gaussian

    Returns
    -------
    cov : numpy.ndarray
        The sum of each Gaussian kernel, or the local covariance kernel
    """
    # Set up the metric and mesh grid
    met = C.c_kms / mu * np.abs(wave - mu)
    x, y = np.meshgrid(met, met)
    r_tap = np.max([x, y], axis=0)
    r2 = x**2 + y**2
    r0 = 4 * sigma

    # Calculate the kernel. Use masking to keep sparse-ish calculations
    kernel = np.zeros((len(x), len(y)))
    mask = r_tap <= r0
    taper = 0.5 + 0.5 * np.cos(np.pi * r_tap[mask] / r0)
    kernel[mask] = taper * amplitude * np.exp(-0.5 * r2[mask] / sigma**2)
    return kernel



import logging

import numpy as np
import scipy.linalg as sl
from scipy.special import loggamma

log = logging.getLogger(__name__)


def get_w_hat(eigenspectra, fluxes):
    """
    Since we will overflow memory if we actually calculate Phi, we have to
    determine w_hat in a memory-efficient manner.

    """
    m = len(eigenspectra)
    M = len(fluxes)
    out = np.empty((M * m,))
    for i in range(m):
        for j in range(M):
            out[i * M + j] = eigenspectra[i].T @ fluxes[j]

    phi_squared = get_phi_squared(eigenspectra, M)
    fac = sl.cho_factor(phi_squared, check_finite=False, overwrite_a=True)
    return sl.cho_solve(fac, out, check_finite=False)


def get_phi_squared(eigenspectra, M):
    """
    Compute Phi.T.dot(Phi) in a memory efficient manner.

    eigenspectra is a list of 1D numpy arrays.
    """
    m = len(eigenspectra)
    out = np.zeros((m * M, m * M))

    # Compute all of the dot products pairwise, beforehand
    dots = np.empty((m, m))
    for i in range(m):
        for j in range(m):
            dots[i, j] = eigenspectra[i].T @ eigenspectra[j]

    for i in range(M * m):
        for jj in range(m):
            ii = i // M
            j = jj * M + (i % M)
            out[i, j] = dots[ii, jj]
    return out


def get_altered_prior_factors(eigenspectra, fluxes):
    """
    Compute the altered priors for the :math:`\\lambda_\\xi` term as in eqns. A24 and
    A25 of Czekala et al. 2015.

    Parameters
    ----------
    eigenspectra : numpy.ndarray
        The PCA eigenspectra
    fluxes : numpy.ndarray
        The vertically stacked input spectra

    Returns
    -------
    """
    w_hat = get_w_hat(eigenspectra, fluxes)
    M, npix = fluxes.shape
    m = len(eigenspectra)

    Phi_w_hat = np.empty((M * npix, 1))
    for i in range(M):
        loss_per_M = np.zeros(npix)
        for j in range(m):
            loss_per_M += eigenspectra[j] * w_hat[i + j * M]
        indices = slice(i * npix, (i + 1) * npix)
        Phi_w_hat[indices] = loss_per_M.reshape(npix, -1)

    F = fluxes.ravel()
    a_prime = 0.5 * M * (npix - m)
    b_prime = 0.5 * (F.T @ F - F.T @ Phi_w_hat)

    return a_prime, b_prime[0]


class Gamma:
    def __init__(self, alpha, beta=1):
        self.alpha = alpha
        self.beta = beta

    def logpdf(self, x):
        lp = (
            self.alpha * np.log(self.beta)
            - loggamma(self.alpha)
            + (self.alpha - 1) * np.log(x)
            - self.beta * x
        )
        return lp

    def pdf(self, x):
        return np.exp(self.logpdf(x))



import logging
import os
import warnings
from typing import Sequence, Optional, Union, Tuple

import h5py
import numpy as np
from nptyping import NDArray
from scipy.interpolate import LinearNDInterpolator
from scipy.linalg import cho_factor, cho_solve
from scipy.optimize import minimize
from sklearn.decomposition import PCA

from Starfish.grid_tools import HDF5Interface
from Starfish.grid_tools.utils import determine_chunk_log
from Starfish.utils import calculate_dv
from .kernels import batch_kernel
from ._utils import get_phi_squared, get_w_hat

log = logging.getLogger(__name__)


class Emulator:
    """
    A Bayesian spectral emulator.

    This emulator offers an interface to spectral libraries that offers interpolation
    while providing a variance-covariance matrix that can be forward-propagated in
    likelihood calculations. For more details, see the appendix from Czekala et al.
    (2015).

    Parameters
    ----------
    grid_points : numpy.ndarray
        The parameter space from the library.
    param_names : array-like of str
        The names of each parameter from the grid
    wavelength : numpy.ndarray
        The wavelength of the library models
    weights : numpy.ndarray
        The PCA weights for the original grid points
    eigenspectra : numpy.ndarray
        The PCA components from the decomposition
    w_hat : numpy.ndarray
        The best-fit weights estimator
    flux_mean : numpy.ndarray
        The mean flux spectrum
    flux_std : numpy.ndarray
        The standard deviation flux spectrum
    lambda_xi : float, optional
        The scaling parameter for the augmented covariance calculations, default is 1
    variances : numpy.ndarray, optional
        The variance parameters for each of Gaussian process, default is array of 1s
    lengthscales : numpy.ndarray, optional
        The lengthscales for each Gaussian process, each row should have length equal
        to number of library parameters, default is arrays of 3 * the max grid
        separation for the grid_points
    name : str, optional
        If provided, will give a name to the emulator; useful for keeping track of
        filenames. Default is None.


    Attributes
    ----------
    params : dict
        The underlying hyperparameter dictionary
    """

    def __init__(
        self,
        grid_points: NDArray[float],
        param_names: Sequence[str],
        wavelength: NDArray[float],
        weights: NDArray[float],
        eigenspectra: NDArray[float],
        w_hat: NDArray[float],
        flux_mean: NDArray[float],
        flux_std: NDArray[float],
        factors: NDArray[float],
        lambda_xi: float = 1.0,
        variances: Optional[NDArray[float]] = None,
        lengthscales: Optional[NDArray[float]] = None,
        name: Optional[str] = None,
    ):
        self.log = logging.getLogger(self.__class__.__name__)
        self.grid_points = grid_points
        self.param_names = param_names
        self.wl = wavelength
        self.weights = weights
        self.eigenspectra = eigenspectra
        self.flux_mean = flux_mean
        self.flux_std = flux_std
        self.factors = factors
        self.factor_interpolator = LinearNDInterpolator(
            grid_points, factors, rescale=True
        )

        self.dv = calculate_dv(wavelength)
        self.ncomps = eigenspectra.shape[0]

        self.hyperparams = {}
        self.name = name

        self.lambda_xi = lambda_xi

        self.variances = (
            variances if variances is not None else 1e4 * np.ones(self.ncomps)
        )

        unique = [sorted(np.unique(param_set)) for param_set in self.grid_points.T]
        self._grid_sep = np.array([np.diff(param).max() for param in unique])

        if lengthscales is None:
            lengthscales = np.tile(3 * self._grid_sep, (self.ncomps, 1))

        self.lengthscales = lengthscales

        # Determine the minimum and maximum bounds of the grid
        self.min_params = grid_points.min(axis=0)
        self.max_params = grid_points.max(axis=0)

        # TODO find better variable names for the following
        self.iPhiPhi = np.linalg.inv(
            get_phi_squared(self.eigenspectra, self.grid_points.shape[0])
        )
        self.v11 = self.iPhiPhi / self.lambda_xi + batch_kernel(
            self.grid_points, self.grid_points, self.variances, self.lengthscales
        )
        self.w_hat = w_hat

        self._trained = False

    @property
    def lambda_xi(self) -> float:
        """
        float : The tuning hyperparameter

        :setter: Sets the value.
        """
        return np.exp(self.hyperparams["log_lambda_xi"])

    @lambda_xi.setter
    def lambda_xi(self, value: float):
        self.hyperparams["log_lambda_xi"] = np.log(value)

    @property
    def variances(self) -> NDArray[float]:
        """
        numpy.ndarray : The variances for each Gaussian process kernel.

        :setter: Sets the variances given an array.
        """
        values = [
            val
            for key, val in self.hyperparams.items()
            if key.startswith("log_variance:")
        ]
        return np.exp(values)

    @variances.setter
    def variances(self, values: NDArray[float]):
        for i, value in enumerate(values):
            self.hyperparams[f"log_variance:{i}"] = np.log(value)

    @property
    def lengthscales(self) -> NDArray[float]:
        """
        numpy.ndarray : The lengthscales for each Gaussian process kernel.

        :setter: Sets the lengthscales given a 2d array
        """
        values = [
            val
            for key, val in self.hyperparams.items()
            if key.startswith("log_lengthscale:")
        ]
        return np.exp(values).reshape(self.ncomps, -1)

    @lengthscales.setter
    def lengthscales(self, values: NDArray[float]):
        for i, value in enumerate(values):
            for j, ls in enumerate(value):
                self.hyperparams[f"log_lengthscale:{i}:{j}"] = np.log(ls)

    def __getitem__(self, key):
        return self.hyperparams[key]

    @classmethod
    def load(cls, filename: Union[str, os.PathLike]):
        """
        Load an emulator from and HDF5 file

        Parameters
        ----------
        filename : str or path-like
        """
        filename = os.path.expandvars(filename)
        with h5py.File(filename, "r") as base:
            grid_points = base["grid_points"][:]
            param_names = base["grid_points"].attrs["names"]
            wavelength = base["wavelength"][:]
            weights = base["weights"][:]
            eigenspectra = base["eigenspectra"][:]
            flux_mean = base["flux_mean"][:]
            flux_std = base["flux_std"][:]
            w_hat = base["w_hat"][:]
            factors = base["factors"][:]
            lambda_xi = base["hyperparameters"]["lambda_xi"][()]
            variances = base["hyperparameters"]["variances"][:]
            lengthscales = base["hyperparameters"]["lengthscales"][:]
            trained = base.attrs["trained"]
            if "name" in base.attrs:
                name = base.attrs["name"]
            else:
                name = ".".join(filename.split(".")[:-1])

        emulator = cls(
            grid_points=grid_points,
            param_names=param_names,
            wavelength=wavelength,
            weights=weights,
            eigenspectra=eigenspectra,
            w_hat=w_hat,
            flux_mean=flux_mean,
            flux_std=flux_std,
            lambda_xi=lambda_xi,
            variances=variances,
            lengthscales=lengthscales,
            name=name,
            factors=factors,
        )
        emulator._trained = trained
        return emulator

    def save(self, filename: Union[str, os.PathLike]):
        """
        Save the emulator to an HDF5 file

        Parameters
        ----------
        filename : str or path-like
        """
        filename = os.path.expandvars(filename)
        with h5py.File(filename, "w") as base:
            grid_points = base.create_dataset(
                "grid_points", data=self.grid_points, compression=9
            )
            grid_points.attrs["names"] = self.param_names
            waves = base.create_dataset("wavelength", data=self.wl, compression=9)
            waves.attrs["units"] = "Angstrom"
            base.create_dataset("weights", data=self.weights, compression=9)
            eigens = base.create_dataset(
                "eigenspectra", data=self.eigenspectra, compression=9
            )
            base.create_dataset("flux_mean", data=self.flux_mean, compression=9)
            base.create_dataset("flux_std", data=self.flux_std, compression=9)
            eigens.attrs["units"] = "erg/cm^2/s/Angstrom"
            base.create_dataset("w_hat", data=self.w_hat, compression=9)
            base.attrs["trained"] = self._trained
            if self.name is not None:
                base.attrs["name"] = self.name
            base.create_dataset("factors", data=self.factors, compression=9)
            hp_group = base.create_group("hyperparameters")
            hp_group.create_dataset("lambda_xi", data=self.lambda_xi)
            hp_group.create_dataset("variances", data=self.variances, compression=9)
            hp_group.create_dataset(
                "lengthscales", data=self.lengthscales, compression=9
            )

        self.log.info("Saved file at {}".format(filename))

    @classmethod
    def from_grid(cls, grid, **pca_kwargs):
        """
        Create an Emulator using PCA decomposition from a GridInterface.

        Parameters
        ----------
        grid : :class:`GridInterface` or str
            The grid interface to decompose
        pca_kwargs : dict, optional
            The keyword arguments to pass to PCA. By default, `n_components=0.99` and
            `svd_solver='full'`.

        See Also
        --------
        sklearn.decomposition.PCA
        """
        # Load grid if a string is given
        if isinstance(grid, str):
            grid = HDF5Interface(grid)

        fluxes = np.array(list(grid.fluxes))
        # Normalize to an average of 1 to remove uninteresting correlation
        norm_factors = fluxes.mean(1)
        fluxes /= norm_factors[:, np.newaxis]
        # Center and whiten
        flux_mean = fluxes.mean(0)
        fluxes -= flux_mean
        flux_std = fluxes.std(0)
        fluxes /= flux_std

        # Perform PCA using sklearn
        default_pca_kwargs = dict(n_components=0.99, svd_solver="full")
        default_pca_kwargs.update(pca_kwargs)
        pca = PCA(**default_pca_kwargs)
        weights = pca.fit_transform(fluxes)
        eigenspectra = pca.components_

        exp_var = pca.explained_variance_ratio_.sum()
        # This is basically the mean square error of the reconstruction
        log.info(
            f"PCA fit {exp_var:.2f}% of the variance with {pca.n_components_:d} components."
        )
        w_hat = get_w_hat(eigenspectra, fluxes)

        emulator = cls(
            grid_points=grid.grid_points,
            param_names=grid.param_names,
            wavelength=grid.wl,
            weights=weights,
            eigenspectra=eigenspectra,
            w_hat=w_hat,
            flux_mean=flux_mean,
            flux_std=flux_std,
            factors=norm_factors,
        )
        return emulator

    def __call__(
        self,
        params: Sequence[float],
        full_cov: bool = True,
        reinterpret_batch: bool = False,
    ) -> Tuple[NDArray[float], NDArray[float]]:
        """
        Gets the mu and cov matrix for a given set of params

        Parameters
        ----------
        params : array_like
            The parameters to sample at. Should be consistent with the shapes of the
            original grid points.
        full_cov : bool, optional
            Return the full covariance or just the variance, default is True. This will
            have no effect of reinterpret_batch is true
        reinterpret_batch : bool, optional
            Will try and return a batch of output matrices if the input params are a
            list of params, default is False.

        Returns
        -------
        mu : numpy.ndarray (len(params),)
        cov : numpy.ndarray (len(params), len(params))

        Raises
        ------
        ValueError
            If full_cov and reinterpret_batch are True
        ValueError
            If querying the emulator outside of its trained grid points
        """
        params = np.atleast_2d(params)

        if full_cov and reinterpret_batch:
            raise ValueError(
                "Cannot reshape the full_covariance matrix for many parameters."
            )

        if not self._trained:
            warnings.warn(
                "This emulator has not been trained and therefore is not reliable. call \
                    emulator.train() to train."
            )

        # If the pars is outside of the range of emulator values, raise a ModelError
        if np.any(params < self.min_params) or np.any(params > self.max_params):
            raise ValueError("Querying emulator outside of original parameter range.")

        # Do this according to R&W eqn 2.18, 2.19
        # Recalculate V12, V21, and V22.
        v12 = batch_kernel(self.grid_points, params, self.variances, self.lengthscales)
        v22 = batch_kernel(params, params, self.variances, self.lengthscales)
        v21 = v12.T

        # Recalculate the covariance
        mu = v21 @ np.linalg.solve(self.v11, self.w_hat)
        cov = v22 - v21 @ np.linalg.solve(self.v11, v12)
        if not full_cov:
            cov = np.diag(cov)
        if reinterpret_batch:
            mu = mu.reshape(-1, self.ncomps, order="F").squeeze()
            cov = cov.reshape(-1, self.ncomps, order="F").squeeze()
        return mu, cov

    @property
    def bulk_fluxes(self) -> NDArray[float]:
        """
        numpy.ndarray: A vertically concatenated vector of the eigenspectra, flux_mean,
        and flux_std (in that order). Used for bulk processing with the emulator.
        """
        return np.vstack([self.eigenspectra, self.flux_mean, self.flux_std])

    def load_flux(
        self, params: Union[Sequence[float], NDArray[float]], norm=False
    ) -> NDArray[float]:
        """
        Interpolate a model given any parameters within the grid's parameter range
        using eigenspectrum reconstruction
        by sampling from the weight distributions.

        Parameters
        ----------
        params : array_like
            The parameters to sample at.

        Returns
        -------
        flux : numpy.ndarray
        """
        mu, cov = self(params, reinterpret_batch=False)
        weights = np.random.multivariate_normal(mu, cov).reshape(-1, self.ncomps)
        X = self.eigenspectra * self.flux_std
        flux = weights @ X + self.flux_mean
        if norm:
            flux *= self.norm_factor(params)[:, np.newaxis]
        return np.squeeze(flux)

    def norm_factor(self, params: Union[Sequence[float], NDArray[float]]) -> float:
        """
        Return the scaling factor for the absolute flux units in flux-normalized spectra

        Parameters
        ----------
        params : array_like
            The parameters to interpolate at

        Returns
        -------
        factor: float
            The multiplicative factor to normalize a spectrum to the model's absolute flux units
        """
        _params = np.asarray(params)
        return self.factor_interpolator(_params)

    def determine_chunk_log(self, wavelength: Sequence[float], buffer: float = 50):
        """
        Possibly truncate the wavelength and eigenspectra in response to some new
        wavelengths

        Parameters
        ----------
        wavelength : array_like
            The new wavelengths to truncate to
        buffer : float, optional
            The wavelength buffer, in Angstrom. Default is 50

        See Also
        --------
        Starfish.grid_tools.utils.determine_chunk_log
        """
        wavelength = np.asarray(wavelength)

        # determine the indices
        wl_min = wavelength.min()
        wl_max = wavelength.max()

        wl_min -= buffer
        wl_max += buffer

        ind = determine_chunk_log(self.wl, wl_min, wl_max)
        trunc_wavelength = self.wl[ind]

        assert (trunc_wavelength.min() <= wl_min) and (
            trunc_wavelength.max() >= wl_max
        ), (
            f"Emulator chunking ({trunc_wavelength.min():.2f}, {trunc_wavelength.max():.2f}) didn't encapsulate "
            f"full wl range ({wl_min:.2f}, {wl_max:.2f})."
        )

        self.wl = trunc_wavelength
        self.eigenspectra = self.eigenspectra[:, ind]

    def train(self, **opt_kwargs):
        """
        Trains the emulator's hyperparameters using gradient descent. This is a light wrapper around `scipy.optimize.minimize`. If you are experiencing problems optimizing the emulator, consider implementing your own training loop, using this function as a template.

        Parameters
        ----------
        **opt_kwargs
            Any arguments to pass to the optimizer. By default, `method='Nelder-Mead'`
            and `maxiter=10000`.

        See Also
        --------
        scipy.optimize.minimize

        """
        # Define our loss function
        def nll(P):
            if np.any(~np.isfinite(P)):
                return np.inf
            self.set_param_vector(P)
            if np.any(self.lengthscales < 2 * self._grid_sep):
                return np.inf
            loss = -self.log_likelihood()
            self.log.debug(f"loss: {loss}")
            return loss

        # Do the optimization
        P0 = self.get_param_vector()

        default_kwargs = {"method": "Nelder-Mead", "options": {"maxiter": 10000}}
        default_kwargs.update(opt_kwargs)
        soln = minimize(nll, P0, **default_kwargs)

        if not soln.success:
            self.log.warning("Optimization did not succeed.")
            self.log.info(soln.message)
        else:
            self.set_param_vector(soln.x)
            self._trained = True
            self.log.info("Finished optimizing emulator hyperparameters")
            self.log.info(self)

    def get_index(self, params: Sequence[float]) -> int:
        """
        Given a list of stellar parameters (corresponding to a grid point),
        deliver the index that corresponds to the
        entry in the fluxes, grid_points, and weights.

        Parameters
        ----------
        params : array_like
            The stellar parameters

        Returns
        -------
        index : int

        """
        params = np.atleast_2d(params)
        marks = np.abs(self.grid_points - np.expand_dims(params, 1)).sum(axis=-1)
        return marks.argmin(axis=1).squeeze()

    def get_param_dict(self) -> dict:
        """
        Gets the dictionary of parameters. This is the same as `Emulator.params`

        Returns
        -------
        dict
        """
        return self.hyperparams

    def set_param_dict(self, params: dict):
        """
        Sets the parameters with a dictionary

        Parameters
        ----------
        params : dict
            The new parameters.
        """
        for key, val in params.items():
            if key in self.hyperparams:
                self.hyperparams[key] = val

        self.v11 = self.iPhiPhi / self.lambda_xi + batch_kernel(
            self.grid_points, self.grid_points, self.variances, self.lengthscales
        )

    def get_param_vector(self) -> NDArray[float]:
        """
        Get a vector of the current trainable parameters of the emulator

        Returns
        -------
        numpy.ndarray
        """
        values = list(self.get_param_dict().values())
        return np.array(values)

    def set_param_vector(self, params: NDArray[float]):
        """
        Set the current trainable parameters given a vector. Must have the same form as
        :meth:`get_param_vector`

        Parameters
        ----------
        params : numpy.ndarray
        """
        parameters = self.get_param_dict()
        if len(params) != len(parameters):
            raise ValueError(
                "params must match length of parameters (get_param_vector())"
            )

        param_dict = dict(zip(self.get_param_dict().keys(), params))
        self.set_param_dict(param_dict)

    def log_likelihood(self) -> float:
        """
        Get the log likelihood of the emulator in its current state as calculated in
        the appendix of Czekala et al. (2015)

        Returns
        -------
        float

        Raises
        ------
        scipy.linalg.LinAlgError
            If the Cholesky factorization fails
        """
        L, flag = cho_factor(self.v11)
        logdet = 2 * np.sum(np.log(np.diag(L)))
        sqmah = self.w_hat @ cho_solve((L, flag), self.w_hat)
        return -(logdet + sqmah) / 2

    def __repr__(self):
        output = "Emulator\n"
        output += "-" * 8 + "\n"
        if self.name is not None:
            output += f"Name: {self.name}\n"
        output += f"Trained: {self._trained}\n"
        output += f"lambda_xi: {self.lambda_xi:.3f}\n"
        output += "Variances:\n"
        output += "\n".join([f"\t{v:.2f}" for v in self.variances])
        output += "\nLengthscales:\n"
        output += "\n".join(
            [
                "\t[ " + " ".join([f"{l:.2f} " for l in ls]) + "]"
                for ls in self.lengthscales
            ]
        )
        output += f"\nLog Likelihood: {self.log_likelihood():.2f}\n"
        return output



import logging

import numpy as np
from nptyping import NDArray
from scipy.optimize import minimize
import scipy.stats as st

import Starfish.constants as C
from .kernels import global_covariance_matrix


def find_residual_peaks(
    model, num_residuals=100, threshold=4.0, buffer=2, wl_range=(0, np.inf)
):
    """
    Find the peaks of the most recent residual and return their properties to aid in
    setting up local kernels

    Parameters
    ----------
    model : Model
        The model to determine peaks from. Need only have a residuals array.
    num_residuals : int, optional
        The number of residuals to average together for determining peaks. By default
        100.
    threshold : float, optional
        The sigma clipping threshold, by default 4.0
    buffer : float, optional
        The minimum distance between peaks, in Angstrom, by default 2.0
    wl_range : 2-tuple
        The (min, max) wavelengths to consider. Default is (0, np.inf)

    Returns
    -------
    means : list
        The means of the found peaks, with the same units as model.data.wave
    """
    residual = np.mean(list(model.residuals)[-num_residuals:], axis=0)
    sigma = residual.std()
    if "global_cov" in model.params:
        ag = np.exp(model.params["global_cov:log_amp"])
        lg = np.exp(model.params["global_cov:log_ls"])
        sigma += global_covariance_matrix(model.data.wave, ag, lg).diagonal()
    mask = np.abs(residual - residual.mean()) > threshold * sigma
    mask &= (model.data.wave > wl_range[0]) & (model.data.wave < wl_range[1])
    peak_waves = model.data.wave[1:-1][mask[1:-1]]
    mus = []
    covered = np.zeros_like(peak_waves, dtype=bool)
    # Sort from largest residual to smallest
    abs_resid = np.abs(residual[1:-1][mask[1:-1]])
    for wl, resid in sorted(
        zip(peak_waves, abs_resid), key=lambda t: t[1], reverse=True
    ):
        if wl in peak_waves[covered]:
            continue
        mus.append(wl)
        ind = (peak_waves >= (wl - buffer)) & (peak_waves <= (wl + buffer))
        covered |= ind

    return mus


def optimize_residual_peaks(model, mus, threshold=0.1, sigma0=50, num_residuals=100):
    """
    Optimize the local covariance parameters based on fitting the residual input means
    as Gaussians around the residuals

    Parameters
    ----------
    model : Model
        The model to determine peaks from. Need only have a residuals array.
    mus : array-like
        The means to instantiate Gaussians at and optimize.
    threshold : float, optional
        This is the threshold for restricting kernels; i.e. if a fit amplitude is less
        than threshold standard deviations then it will be thrown away. Default is 0.1
    sigma0 : float, optional
        The initial standard deviation (in Angstrom) of each Gaussian. Default is 50
        Angstrom.
    num_residuals : int, optional
        The number of residuals to average together for determining peaks. By default
        100.

    Returns
    -------
    dict
        A dictionary of optimized parameters ready to be plugged into model["local_cov"]

    Warning
    -------
    I have had inconsistent results with this optimization, be mindful of your outputs
    and consider hand-tuning after optimizing.
    """
    residual = np.mean(list(model.residuals)[-num_residuals:], axis=0)
    amp_cutoff = threshold * residual.std()
    if "global_cov" in model.params:
        ag = np.exp(model.params["global_cov:log_amp"])
        lg = np.exp(model.params["global_cov:log_ls"])
        global_cov = global_covariance_matrix(model.data.wave, ag, lg)
    else:
        global_cov = None

    def chi2(P, wave, resid, sigma):
        log_amp, mu, log_sigma = P
        _amp = np.exp(log_amp)
        _sigma = np.exp(log_sigma)
        if _amp == 0 or _sigma == 0:
            return np.Inf
        # Logistic prior for the widths out to 2sigma
        prior = st.uniform.logpdf(_sigma, 0, 2 * sigma0)
        # Put prior on widths and heights such that the integrated area should be less
        # than the trapezoidal area of the residual
        area_max = 2 * _sigma * np.abs(resid).max()
        # Area under a gaussian
        # https://en.wikipedia.org/wiki/Gaussian_function#Integral_of_a_Gaussian_function
        area = np.sqrt(2 * np.pi * _sigma) * _amp
        prior += st.uniform.logpdf(area, 0, area_max)
        prior += st.uniform.logpdf(_amp, 0, np.abs(resid).max())
        rr = C.c_kms / mu * np.abs(wave - mu)
        gauss = _amp * np.exp(-0.5 * (rr / _sigma) ** 2)
        R = gauss - resid
        return np.sum((R / sigma) ** 2) - prior

    params = []

    for mu in mus:
        mask = (model.data.wave > mu - sigma0) & (model.data.wave < mu + sigma0)
        wave = model.data.wave[mask]
        resid = residual[mask]
        sigma = model.data.sigma[mask]
        if global_cov is not None:
            sigma += global_cov.diagonal()[mask]

        P0 = np.array([np.log(np.abs(resid).max()), mu, np.log(sigma0)])

        soln = minimize(
            chi2,
            P0,
            args=(wave, resid, sigma),
            method="Nelder-Mead",
            options=dict(maxiter=1000),
        )
        if soln.x[0] > np.log(amp_cutoff):
            params.append(
                {"log_amp": soln.x[0], "mu": soln.x[1], "log_sigma": soln.x[2]}
            )

    return sorted(params, key=lambda s: s["mu"])


log = logging.getLogger(__name__)


def covariance_debugger(cov: NDArray[float]):
    """
    Special debugging information for the covariance matrix decomposition.
    """
    log.info(f"{'Covariance Debugger':-^60}".format())
    log.info("See https://github.com/iancze/Starfish/issues/26")
    log.info("Covariance matrix at a glance:")
    if cov.diagonal().min() < 0.0:
        log.warning("- Negative entries on the diagonal:")
        log.info("\t- Check uncertainty estimates: should all be positive")
    elif np.any(np.isnan(cov.diagonal())):
        log.warning("- Covariance matrix has a NaN value on the diagonal")
    else:
        if not np.allclose(cov, cov.T):
            log.warning("- The covariance matrix is highly asymmetric")

        # Still might have an asymmetric matrix below `allclose` threshold
        eigenvalues, eigenvectors = np.linalg.eigh(cov)
        n_neg = (eigenvalues < 0).sum()
        n_tot = len(eigenvalues)
        log.info(f"- There are {n_neg} negative eigenvalues out of {n_tot}.")

        def mark(val):
            return ">" if val < 0 else "."

        log.info("Covariance matrix eigenvalues:")
        for i in range(10):
            log.info(
                "{: >6} {:{fill}>20.3e}".format(
                    i, eigenvalues[i], fill=mark(eigenvalues[i])
                )
            )
        log.info("{: >15}".format("..."))
        for i in range(10):
            log.info(
                "{: >6} {:{fill}>20.3e}".format(
                    n_tot - 10 + i,
                    eigenvalues[-10 + i],
                    fill=mark(eigenvalues[-10 + i]),
                )
            )

    log.info(f"{'-':-^60}")



from dataclasses import dataclass
from typing import Tuple

# TODO convert to dataclass

# Convert R to FWHM in km/s by \Delta v = c/R
@dataclass
class Instrument:
    """
    Object describing an instrument. This will be used by other methods for
    processing raw synthetic spectra.

    Parameters
    ----------
    name: string
        name of the instrument
    FWHM: float
        the FWHM of the instrumental profile in km/s
    wl_range: Tuple (low, high)
        wavelength range of instrument
    oversampling: float, optional
        how many samples fit across the :attr:`FWHM`. Default is 4.0
    """

    name: str
    FWHM: float
    wl_range: Tuple[float]
    oversampling: float = 4.0

    def __str__(self):
        """
        Prints the relevant properties of the instrument.
        """
        return (
            "instrument Name: {}, FWHM: {:.1f}, oversampling: {:.0f}, "
            "wl_range: {}".format(
                self.name, self.FWHM, self.oversampling, self.wl_range
            )
        )


class TRES(Instrument):
    """TRES instrument"""

    def __init__(self, name="TRES", FWHM=6.8, wl_range=(3500, 9500)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)
        # sets the FWHM and wl_range


class Reticon(Instrument):
    """Reticon instrument"""

    def __init__(self, name="Reticon", FWHM=8.5, wl_range=(5145, 5250)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)


class KPNO(Instrument):
    """KNPO instrument"""

    def __init__(self, name="KPNO", FWHM=14.4, wl_range=(6250, 6650)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)


class SPEX(Instrument):
    """SPEX instrument at IRTF in Hawaii"""

    def __init__(self, name="SPEX", FWHM=150.0, wl_range=(7500, 54000)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)


class SPEX_SXD(SPEX):
    """SPEX instrument at IRTF in Hawaii short mode (reduced wavelength range)"""

    def __init__(self, name="SPEX_SXD"):
        super().__init__(name=name, wl_range=(7500, 26000))


class SPEX_PRISM(SPEX):
    """SPEX instrument at IRTF in Hawaii low-res PRISM mode (reduced wavelength range)"""

    def __init__(self, name="SPEX_SXD"):
        super().__init__(name=name, FWHM=1500, wl_range=(7500, 26000))


class IGRINS(Instrument):
    """IGRINS Instruments Abstract Class"""

    def __init__(self, wl_range, name="IGRINS"):
        super().__init__(name=name, FWHM=7.5, wl_range=wl_range)
        self.air = False


class IGRINS_H(IGRINS):
    """IGRINS H band instrument"""

    def __init__(self, name="IGRINS_H", wl_range=(14250, 18400)):
        super().__init__(name=name, wl_range=wl_range)


class IGRINS_K(IGRINS):
    """IGRINS K band instrument"""

    def __init__(self, name="IGRINS_K", wl_range=(18500, 25200)):
        super().__init__(name=name, wl_range=wl_range)


class ESPaDOnS(Instrument):
    """ESPaDOnS instrument"""

    def __init__(self, name="ESPaDOnS", FWHM=4.4, wl_range=(3700, 10500)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)


class DCT_DeVeny(Instrument):
    """DCT DeVeny spectrograph instrument."""

    def __init__(self, name="DCT_DeVeny", FWHM=105.2, wl_range=(6000, 10000)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)


class WIYN_Hydra(Instrument):
    """WIYN Hydra spectrograph instrument."""

    def __init__(self, name="WIYN_Hydra", FWHM=300.0, wl_range=(5500, 10500)):
        super().__init__(name=name, FWHM=FWHM, wl_range=wl_range)



__version__ = "0.4.2"

from .spectrum import Spectrum

__all__ = [
    "constants",
    "emulator",
    "grid_tools",
    "models",
    "samplers",
    "spectrum",
    "Spectrum",
    "transforms",
    "utils",
]



from typing import Sequence

import numpy as np

import Starfish.constants as C


def calculate_dv(wave: Sequence):
    """
    Given a wavelength array, calculate the minimum ``dv`` of the array.

    Parameters
    ----------
    wave : array-like
        The wavelength array

    Returns
    -------
    float
        delta-v in units of km/s
    """
    return C.c_kms * np.min(np.diff(wave) / wave[:-1])


def calculate_dv_dict(wave_dict):
    """
    Given a ``wave_dict``, calculate the velocity spacing.

    Parameters
    ---------
    wave_dict : dict
        wavelength dictionary

    Returns
    -------
    float
        delta-v in units of km/s
    """
    CDELT1 = wave_dict["CDELT1"]
    dv = C.c_kms * (10**CDELT1 - 1)
    return dv


def create_log_lam_grid(dv, start, end):
    """
    Create a log lambda spaced grid with ``N_points`` equal to a power of 2 for
    ease of FFT.

    Parameters
    ----------
    dv : float
        Upper bound on the velocity spacing in km/s
    start : float
        starting wavelength (inclusive) in Angstrom
    end : float
        ending wavelength (inclusive) in Angstrom

    Returns
    -------
    dict
        a wavelength dictionary containing the specified properties. Note that the returned dv will be less than or equal to the specified dv.

    Raises
    ------
    ValueError
        If starting wavelength is not less than ending wavelength
    ValueError
        If any of the wavelengths are less than 0
    """
    if start >= end:
        raise ValueError("Wavelength must be increasing, but start >= end")

    if start <= 0 or end <= 0:
        raise ValueError("Cannot have negative or 0 wavelength")

    CDELT_temp = np.log10(dv / C.c_kms + 1.0)
    CRVAL1 = np.log10(start)
    CRVALN = np.log10(end)
    N = (CRVALN - CRVAL1) / CDELT_temp
    NAXIS1 = 2
    while NAXIS1 < N:  # Make NAXIS1 an integer power of 2 for FFT purposes
        NAXIS1 *= 2

    CDELT1 = (CRVALN - CRVAL1) / (NAXIS1 - 1)

    p = np.arange(NAXIS1)
    wl = 10 ** (CRVAL1 + CDELT1 * p)
    return {"wl": wl, "CRVAL1": CRVAL1, "CDELT1": CDELT1, "NAXIS1": NAXIS1}



import itertools
from collections import OrderedDict
import logging

import numpy as np
from scipy.interpolate import interp1d

from .utils import determine_chunk_log

BUFFER = 50


class IndexInterpolator:
    """
    Object to return fractional distance between grid points of a single grid variable.

    :param parameter_list: list of parameter values
    :type parameter_list: iterable
    """

    def __init__(self, parameter_list):
        parameter_list = list(parameter_list)
        self.npars = len(parameter_list)
        self.parameter_list = [np.unique(pars) for pars in parameter_list]
        idxs = [np.arange(len(pars)) for pars in self.parameter_list]
        self.index_interpolators = [
            interp1d(pars, idx, kind="linear")
            for pars, idx in zip(self.parameter_list, idxs)
        ]

    def __call__(self, param):
        """
        Evaluate the interpolator at a parameter.

        :param param:
        :type param: list
        :raises ValueError: if *value* is out of bounds.

        :returns: ((low_val, high_val), (low_dist, high_dist)), the lower and higher bounding points in the grid
        and the fractional distance (0 - 1) from the two points.
        """
        if len(param) != self.npars:
            raise ValueError(
                "Incorrect number of parameters. Expected {} but got {}".format(
                    self.npars, len(param)
                )
            )
        lows = np.empty(self.npars)
        highs = np.empty(self.npars)
        fracs = np.empty(self.npars)
        for i in range(self.npars):
            # get interpolated index
            try:
                index = self.index_interpolators[i](param[i])
            except ValueError:
                raise ValueError("Requested param {} is out of bounds.".format(param))
            low = np.floor(index).astype(int)
            high = np.ceil(index).astype(int)
            frac = index - low
            # get bounding params
            lows[i] = self.parameter_list[i][low]
            highs[i] = self.parameter_list[i][high]
            fracs[i] = frac

        return (lows, highs), (1 - fracs, fracs)


class Interpolator:
    """
    Quickly and efficiently interpolate a synthetic spectrum for use in an MCMC
    simulation. Caches spectra for easier memory load.

    :param interface: The interface to the spectra
    :type interface: :obj:`HDF5Interface` (recommended) or :obj:`RawGridInterface`
    :param wl_range: If provided, the data wavelength range of the region you are trying to fit. Used to truncate the
        grid for speed. Default is (0, np.inf)
    :type wl_range: tuple (min, max)
    :param cache_max: maximum number of spectra to hold in cache
    :type cache_max: int
    :param cache_dump: how many spectra to purge from the cache once :attr:`cache_max` is reached
    :type cache_dump: int

    .. warning:: Interpolation causes degradation of information of the model spectra without properly forward
        propagating the errors from interpolation. We highly recommend using the :ref:`Spectral Emulator <Spectral
        Emulator>`
    """

    def __init__(self, interface, wl_range=(0, np.inf), cache_max=256, cache_dump=64):

        self.interface = interface

        self.wl = self.interface.wl
        mask = (self.wl < wl_range[-1]) & (self.wl < wl_range[1])
        self.wl = self.wl[mask]
        self.dv = self.interface.dv
        self.npars = len(interface.param_names)
        self._determine_chunk_log()

        self._setup_index_interpolators()
        self.cache = OrderedDict([])
        self.cache_max = cache_max
        self.cache_dump = (
            cache_dump  # how many to clear once the maximum cache has been reached
        )
        self.log = logging.getLogger(self.__class__.__name__)

    def _determine_chunk_log(self):
        """
        Determine the minimum chunk size that we can use and then
        truncate the synthetic wavelength grid and the returned spectra.

        Assumes HDF5Interface is LogLambda spaced, because otherwise you shouldn't need a grid
        with 2^n points, because you would need to interpolate in wl space after this anyway.
        """

        wl_interface = self.interface.wl  # The grid we will be truncating.
        wl_min, wl_max = np.min(self.wl), np.max(self.wl)

        # Previously this routine retuned a tuple () which was the ranges to truncate to.
        # Now this routine returns a Boolean mask,
        # so we need to go and find the first and last true values
        ind = determine_chunk_log(wl_interface, wl_min, wl_max)
        self.wl = self.wl[ind]

        # Find the index of the first and last true values
        self.interface.ind = np.argwhere(ind)[0][0], np.argwhere(ind)[-1][0] + 1

    def __call__(self, parameters):
        """
        Interpolate a spectrum

        :param parameters: stellar parameters
        :type parameters: numpy.ndarray or list

        .. note:: Automatically pops :attr:`cache_dump` items from cache if full.
        """
        parameters = np.asarray(parameters)

        if len(self.cache) > self.cache_max:
            [self.cache.popitem(False) for i in range(self.cache_dump)]
            self.cache_counter = 0

        return self.interpolate(parameters)

    def _setup_index_interpolators(self):
        # create an interpolator between grid points indices.
        # Given a parameter value, produce fractional index between two points
        # Store the interpolators as a list
        self.index_interpolator = IndexInterpolator(self.interface.points)

        lenF = self.interface.ind[1] - self.interface.ind[0]
        self.fluxes = np.empty((2**self.npars, lenF))  # 8 rows, for temp, logg, Z

    def interpolate(self, parameters):
        """
        Interpolate a spectrum without clearing cache. Recommended to use :meth:`__call__` instead to
        take advantage of caching.

        :param parameters: grid parameters
        :type parameters: numpy.ndarray or list

        :raises ValueError: if parameters are out of bounds.

        """
        parameters = np.asarray(parameters)

        # Previously, parameters was a dictionary of the stellar parameters.
        # Now that we have moved over to arrays, it is a numpy array.

        params, weights = self.index_interpolator(parameters)

        # Selects all the possible combinations of parameters and weights
        param_combos = list(itertools.product(*np.array(params).T))
        weight_combos = list(itertools.product(*np.array(weights).T))

        # Assemble key list necessary for indexing cache
        key_list = [self.interface.key_name.format(*param) for param in param_combos]
        weight_list = np.array([np.prod(weight) for weight in weight_combos])

        assert np.allclose(
            np.sum(weight_list), np.array(1.0)
        ), "Sum of weights must equal 1, {}".format(np.sum(weight_list))

        # Assemble flux vector from cache, or load into cache if not there
        for i, param in enumerate(param_combos):
            key = key_list[i]
            if key not in self.cache.keys():
                # This method already allows loading only the relevant region from HDF5
                fl = self.interface.load_flux(param, header=False)
                self.cache[key] = fl

            self.fluxes[i] = self.cache[key] * weight_list[i]

        return np.sum(self.fluxes, axis=0)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import os
import re

version = ""

with open(os.path.join("Starfish", "__init__.py"), "r") as fh:
    for line in fh.readlines():
        m = re.search("__version__ = [\"'](.+)[\"']", line)
        if m:
            version = m.group(1)


with open("README.md", "r") as fh:
    readme = fh.read()

setup(
    long_description=readme,
    long_description_content_type="text/markdown",
    name="astrostarfish",
    version=version,
    description="Covariance tools for fitting stellar spectra",
    python_requires=">=3.7,<3.10",
    project_urls={
        "repository": "https://github.com/iancze/Starfish",
        "documentation": "https://starfish.rtfd.io",
    },
    author="Ian Czekala",
    author_email="iancze@gmail.com",
    maintainer="Miles Lucas <mdlucas@hawaii.edu",
    license="BSD-4-Clause",
    keywords="Science Astronomy Physics Data Science",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_data={},
    install_requires=[
        "astropy>=4,<6",
        "dataclasses>=0.6,<0.7",
        "extinction==0.4.*",
        "flatdict==4.*",
        "h5py==3.*",
        "nptyping==1.*",
        "numpy>=1.16,<2",
        "scikit-learn>=0.22,<2",
        "scipy>=1.3.0,<2",
        "toml>=0.10.1,<0.11",
        "tqdm==4.*",
    ],
    extras_require={
        "docs": [
            "nbsphinx==0.*,>=0.4.2",
            "pandoc==2.*",
            "sphinx==2.*",
            "sphinx-bootstrap-theme==0.*,>=0.7.1",
            "IPython",
            "sphinx-autodoc-typehints==1.10.3",
        ],
        "test": [
            "coveralls==1.*,>=1.8.0",
            "pytest==4.*,>=4.6.0",
            "pytest-black",
            "pytest-cov==2.*,>=2.7.0",
            "pytest-benchmark",
        ],
    },
)



import h5py
import numpy as np
from dataclasses import dataclass
from nptyping import NDArray
from typing import Optional


@dataclass
class Order:
    """
    A data class to hold astronomical spectra orders.

    Parameters
    ----------
    _wave : numpy.ndarray
        The full wavelength array
    _flux : numpy.ndarray
        The full flux array
    _sigma : numpy.ndarray, optional
        The full sigma array. If None, will default to all 0s. Default is None
    mask : numpy.ndarray, optional
        The full mask. If None, will default to all Trues. Default is None

    Attributes
    ----------
    name : str
    """

    _wave: NDArray[float]
    _flux: NDArray[float]
    _sigma: Optional[NDArray[float]] = None
    mask: Optional[NDArray[bool]] = None

    def __post_init__(self):
        if self._sigma is None:
            self._sigma = np.zeros_like(self._flux)
        if self.mask is None:
            self.mask = np.ones_like(self._wave, dtype=bool)

    @property
    def wave(self):
        """
        numpy.ndarray : The masked wavelength array
        """
        return self._wave[self.mask]

    @property
    def flux(self):
        """
        numpy.ndarray : The masked flux array
        """
        return self._flux[self.mask]

    @property
    def sigma(self):
        """
        numpy.ndarray : The masked flux uncertainty array
        """
        return self._sigma[self.mask]

    def __len__(self):
        return len(self._wave)


class Spectrum:
    """
    Object to store astronomical spectra.

    Parameters
    ----------
    waves : 1D or 2D array-like
        wavelength in Angtsrom
    fluxes : 1D or 2D array-like
         flux (in f_lam)
    sigmas : 1D or 2D array-like, optional
        Poisson noise (in f_lam). If not specified, will be zeros. Default is None
    masks : 1D or 2D array-like, optional
        Mask to blot out bad pixels or emission regions. Must be castable to boolean. If None, will create a mask of all True. Default is None
    name : str, optional
        The name of this spectrum. Default is "Spectrum"

    Note
    ----
    If the waves, fluxes, and sigmas are provided as 1D arrays (say for a single order), they will be converted to 2D arrays with length 1 in the 0-axis.

    Warning
    -------
    For now, the Spectrum waves, fluxes, sigmas, and masks must be a rectangular grid. No ragged Echelle orders allowed.

    Attributes
    ----------
    name : str
        The name of the spectrum
    """

    def __init__(self, waves, fluxes, sigmas=None, masks=None, name="Spectrum"):
        waves = np.atleast_2d(waves)
        fluxes = np.atleast_2d(fluxes)

        if sigmas is not None:
            sigmas = np.atleast_2d(sigmas)
        else:
            sigmas = np.ones_like(fluxes)

        if masks is not None:
            masks = np.atleast_2d(masks).astype(bool)
        else:
            masks = np.ones_like(waves, dtype=bool)
        assert fluxes.shape == waves.shape, "flux array incompatible shape."
        assert sigmas.shape == waves.shape, "sigma array incompatible shape."
        assert masks.shape == waves.shape, "mask array incompatible shape."
        self.orders = []
        for i in range(len(waves)):
            self.orders.append(Order(waves[i], fluxes[i], sigmas[i], masks[i]))
        self.name = name

    def __getitem__(self, index: int):
        return self.orders[index]

    def __setitem__(self, index: int, order: Order):
        if len(order) != len(self.orders[0]):
            raise ValueError("Invalid order length; no ragged spectra allowed")
        self.orders[index] = order

    def __len__(self):
        return len(self.orders)

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self.orders):
            n, self._n = self._n, self._n + 1
            return self.orders[n]
        else:
            raise StopIteration

    # Masked properties
    @property
    def waves(self) -> np.ndarray:
        """
        numpy.ndarray : The 2 dimensional masked wavelength arrays
        """
        waves = [o.wave for o in self.orders]
        return np.asarray(waves)

    @property
    def fluxes(self) -> np.ndarray:
        """
        numpy.ndarray : The 2 dimensional masked flux arrays
        """
        fluxes = [o.flux for o in self.orders]
        return np.asarray(fluxes)

    @property
    def sigmas(self) -> np.ndarray:
        """
        numpy.ndarray : The 2 dimensional masked flux uncertainty arrays
        """
        sigmas = [o.sigma for o in self.orders]
        return np.asarray(sigmas)

    # Unmasked properties
    @property
    def _waves(self) -> np.ndarray:
        _waves = [o._wave for o in self.orders]
        return np.asarray(_waves)

    @property
    def _fluxes(self) -> np.ndarray:
        _fluxes = [o._flux for o in self.orders]
        return np.asarray(_fluxes)

    @property
    def _sigmas(self) -> np.ndarray:
        _sigmas = [o._sigma for o in self.orders]
        return np.asarray(_sigmas)

    @property
    def masks(self) -> np.ndarray:
        """
        np.ndarray: The full 2-dimensional boolean masks
        """
        waves = [o.wave for o in self.orders]
        return np.asarray(waves)

    @property
    def shape(self):
        """
        numpy.ndarray: The shape of the spectrum, *(norders, npixels)*

        :setter: Tries to reshape the data into a new arrangement of orders and pixels following numpy reshaping rules.
        """
        return (len(self), len(self.orders[0]))

    @shape.setter
    def shape(self, shape):
        new = self.reshape(shape)
        self.__dict__.update(new.__dict__)

    def reshape(self, shape):
        """
        Reshape the spectrum to the new shape. Obeys the same rules that numpy reshaping does. Note this is not done in-place.

        Parameters
        ----------
        shape : tuple
            The new shape of the spectrum. Must abide by numpy reshaping rules.

        Returns
        -------
        Spectrum
            The reshaped spectrum
        """
        waves = self._waves.reshape(shape)
        fluxes = self._fluxes.reshape(shape)
        sigmas = self._sigmas.reshape(shape)
        masks = self.masks.reshape(shape)
        return self.__class__(waves, fluxes, sigmas, masks, name=self.name)

    @classmethod
    def load(cls, filename):
        """
        Load a spectrum from an hdf5 file

        Parameters
        ----------
        filename : str or path-like
            The path to the HDF5 file.

        See Also
        --------
        :meth:`save`
        """
        with h5py.File(filename, "r") as base:
            if "name" in base.attrs:
                name = base.attrs["name"]
            else:
                name = None
            waves = base["waves"][:]
            fluxes = base["fluxes"][:]
            sigmas = base["sigmas"][:]
            masks = base["masks"][:]
        return cls(waves, fluxes, sigmas, masks, name=name)

    def save(self, filename):
        """
        Takes the current DataSpectrum and writes it to an HDF5 file.

        Parameters
        ----------
        filename: str or path-like
            The filename to write to. Will not create any missing directories.

        See Also
        --------
        :meth:`load`
        """

        with h5py.File(filename, "w") as base:
            base.create_dataset("waves", data=self.waves, compression=9)
            base.create_dataset("fluxes", data=self.fluxes, compression=9)
            base.create_dataset("sigmas", data=self.sigmas, compression=9)
            base.create_dataset("masks", data=self.masks, compression=9)
            if self.name is not None:
                base.attrs["name"] = self.name

    def plot(self, ax=None, **kwargs):
        """
        Plot all the orders of the spectrum

        Parameters
        ----------
        ax : matplotlib.Axes, optional
            If provided, will plot on this axis. Otherwise, will create a new axis, by
            default None

        Returns
        -------
        matplotlib.Axes
            The axis that was plotted on
        """
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots(1, 1)

        plot_params = {"lw": 0.5}
        plot_params.update(kwargs)
        # Plot orders
        for i, order in enumerate(self.orders):
            ax.plot(order._wave, order._flux, label=f"Order: {i}", **plot_params)

        # Now plot masks
        ylims = ax.get_ylim()
        for i, order in enumerate(self.orders):
            ax.fill_between(
                order._wave,
                *ylims,
                color="k",
                alpha=0.1,
                where=~order.mask,
                label="Mask" if i == 0 else None,
            )

        ax.set_yscale("log")
        ax.set_ylabel(r"$f_\lambda$ [$erg/cm^2/s/cm$]")
        ax.set_xlabel(r"$\lambda$ [$\AA$]")
        ax.legend()
        if self.name is not None:
            ax.set_title(self.name)
        fig.tight_layout()
        return ax

    def __repr__(self):
        return f"{self.name} ({self.waves.shape[0]} orders)"



import os
import logging
import tqdm
from urllib.request import urlretrieve, URLError
import multiprocessing as mp
import itertools

import numpy as np


log = logging.getLogger(__name__)


def download_PHOENIX_models(path, ranges=None, parameters=None):
    """
    Download the PHOENIX grid models from the Goettingen servers. This will skip over
    any ill-defined files or any files that already exist on disk in the given folder.

    Parameters
    ----------
    path : str or path-like
        The base directory to save the files in.
    ranges : iterable of (min, max), optional
        Each entry in ranges should be (min, max) for the associated parameter, in the
        order [Teff, logg, Z, (Alpha)]. Cannot be used with :attr:`parameters`. Default
        is None
    parameters : iterable of iterables of length 3 or length 4, optional
        The parameters to download. Should be a list of parameters where parameters can
        either be [Teff, logg, Z] or [Teff, logg, Z, Alpha]. All values should be
        floats or integers and not string. If no value provided, will download all
        models. Default is None

    Raises
    ------
    ValueError
        If both ``parameters`` and ``ranges`` are specified

    Warning
    -------
    This will create any directories if they do not exist

    Warning
    -------
    Please use this responsibly to avoid over-saturating the connection to the
    Gottingen servers.

    Examples
    --------

    .. code-block:: python

        from Starfish.grid_tools import download_PHOENIX_models

        ranges = [
            [5000, 5200] # T
            [4.0, 5.0] # logg
            [0, 0] # Z
        ]
        download_PHOENIX_models(path='models', ranges=ranges)

    or equivalently using ``parameters`` syntax

    .. code-block:: python

        from itertools import product
        from Starfish.grid_tools import download_PHOENIX_models

        T = [6000, 6100, 6200]
        logg = [4.0, 4.5, 5.0]
        Z = [0]
        params = product(T, logg, Z)
        download_PHOENIX_models(path='models', parameters=params)

    """
    if parameters is not None and ranges is not None:
        raise ValueError("Cannot specify both 'parameters' and 'ranges'")

    wave_url = "http://phoenix.astro.physik.uni-goettingen.de/data/HiResFITS/WAVE_PHOENIX-ACES-AGSS-COND-2011.fits"
    wave_file = os.path.join(path, "WAVE_PHOENIX-ACES-AGSS-COND-2011.fits")
    flux_file_formatter = (
        "http://phoenix.astro.physik.uni-goettingen.de/data/HiResFITS/PHOENIX-ACES-AGSS-COND-2011"
        "/Z{2:s}{3:s}/lte{0:05.0f}-{1:03.2f}{2:s}{3:s}.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits"
    )
    output_formatter = "Z{2:s}{3:s}/lte{0:05.0f}-{1:03.2f}{2:s}{3:s}.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits"

    os.makedirs(path, exist_ok=True)
    # Download step
    log.info("Starting Download of PHOENIX ACES models to {}".format(path))
    if not os.path.exists(wave_file):
        log.info("Downloading wavelength file")
        urlretrieve(wave_url, wave_file)

    # Have to wait until wavelength file is downloaded before importing
    from .interfaces import PHOENIXGridInterface, PHOENIXGridInterfaceNoAlpha

    # Kind of messy, sorry
    if parameters is None:
        if ranges is not None:
            if len(ranges) == 3:
                grid = PHOENIXGridInterfaceNoAlpha(path)
            elif len(ranges) == 4:
                grid = PHOENIXGridInterface(path)
        parameters = list(itertools.product(*grid.points))
    elif len(parameters[0]) == 3:
        grid = PHOENIXGridInterfaceNoAlpha(path)
    elif len(parameters[0]) == 4:
        grid = PHOENIXGridInterface(path)

    if ranges is not None:
        _ranges = np.asarray(ranges)
        min_params = _ranges.T[0]
        max_params = _ranges.T[1]
    else:
        min_params = np.tile(-np.inf, len(parameters[0]))
        max_params = np.tile(np.inf, len(parameters[0]))

    # I hate to iterate here, but this way the progress bar doesn't show something like
    # 7000 parameters and skips thousands at a time
    params = []
    for p in parameters:
        if np.all(p >= min_params) and np.all(p <= max_params):
            params.append(p)

    pbar = tqdm.tqdm(params)
    for p in pbar:
        # Skip irregularities from grid
        try:
            grid.check_params(p)
        except ValueError:
            continue

        tmp_p = [p[0], p[1]]
        # Create the Z string. Have to do this because PHOENIX models use - sign for 0.0
        Zstr = "-0.0" if p[2] == 0 else "{:+.1f}".format(p[2])
        tmp_p.append(Zstr)
        # Create the Alpha string, which is nothing if alpha is 0 or unspecified
        if len(p) == 4:
            Astr = "" if p[3] == 0 else ".Alpha={:+.2f}".format(p[3])
        else:
            Astr = ""
        tmp_p.append(Astr)
        url = flux_file_formatter.format(*tmp_p)
        pbar.set_description(url.split("/")[-1])
        output_file = os.path.join(path, output_formatter.format(*tmp_p))
        if not os.path.exists(output_file):
            os.makedirs(os.path.dirname(output_file), exist_ok=True)
            try:
                urlretrieve(url, output_file)
            except URLError:
                log.warning(
                    f"Parameters {p} not found. Double check they are on PHOENIX grid"
                )


def chunk_list(mylist, n=mp.cpu_count()):
    """
    Divide a lengthy parameter list into chunks for parallel processing and
    backfill if necessary.

    :param mylist: a lengthy list of parameter combinations
    :type mylist: 1-D list
    :param n: number of chunks to divide list into. Default is ``mp.cpu_count()``
    :type n: integer

    :returns: **chunks** (*2-D list* of shape (n, -1)) a list of chunked parameter lists.

    """
    if isinstance(mylist, np.ndarray):
        mylist = list(mylist)
    length = len(mylist)
    size = int(length / n)
    # fill with evenly divisible
    chunks = [mylist[0 + size * i : size * (i + 1)] for i in range(n)]
    leftover = length - size * n
    edge = size * n
    for i in range(leftover):  # backfill each with the last item
        chunks[i % n].append(mylist[edge + i])
    return chunks


def determine_chunk_log(wl, wl_min, wl_max):
    """
    Take in a wavelength array and then, given two minimum bounds, determine
    the boolean indices that will allow us to truncate this grid to near the
    requested bounds while forcing the wl length to be a power of 2.

    :param wl: wavelength array
    :type wl: np.ndarray
    :param wl_min: minimum required wavelength
    :type wl_min: float
    :param wl_max: maximum required wavelength
    :type wl_max: float

    :returns: numpy.ndarray boolean array

    """

    # wl_min and wl_max must of course be within the bounds of wl
    assert wl_min >= np.min(wl) and wl_max <= np.max(
        wl
    ), "determine_chunk_log: wl_min {:.2f} and wl_max {:.2f} are not within the bounds of the grid {:.2f} to {:.2f}.".format(
        wl_min, wl_max, np.min(wl), np.max(wl)
    )

    # Find the smallest length synthetic spectrum that is a power of 2 in length
    # and longer than the number of points contained between wl_min and wl_max
    len_wl = len(wl)
    npoints = np.sum((wl >= wl_min) & (wl <= wl_max))
    chunk = len_wl
    inds = (0, chunk)

    # This loop will exit with chunk being the smallest power of 2 that is
    # larger than npoints
    while chunk > npoints:
        if chunk / 2 > npoints:
            chunk = chunk // 2
        else:
            break

    assert type(chunk) == int, "Chunk is not an integer!. Chunk is {}".format(chunk)

    if chunk < len_wl:
        # Now that we have determined the length of the chunk of the synthetic
        # spectrum, determine indices that straddle the data spectrum.

        # Find the index that corresponds to the wl at the center of the data spectrum
        center_wl = (wl_min + wl_max) / 2.0
        center_ind = (np.abs(wl - center_wl)).argmin()

        # Take a chunk that straddles either side.
        inds = (center_ind - chunk // 2, center_ind + chunk // 2)

        ind = (np.arange(len_wl) >= inds[0]) & (np.arange(len_wl) < inds[1])
    else:
        print("keeping grid as is")
        ind = np.ones_like(wl, dtype="bool")

    assert (min(wl[ind]) <= wl_min) and (max(wl[ind]) >= wl_max), (
        "Model"
        "Interpolator chunking ({:.2f}, {:.2f}) didn't encapsulate full"
        " wl range ({:.2f}, {:.2f}).".format(min(wl[ind]), max(wl[ind]), wl_min, wl_max)
    )

    return ind


def vacuum_to_air(wl):
    """
    Converts vacuum wavelengths to air wavelengths using the Ciddor 1996 formula.

    :param wl: input vacuum wavelengths
    :type wl: numpy.ndarray

    :returns: numpy.ndarray

    .. note::

        CA Prieto recommends this as more accurate than the IAU standard.

    """
    wl = np.asarray(wl)

    sigma = (1e4 / wl) ** 2
    f = 1.0 + 0.05792105 / (238.0185 - sigma) + 0.00167917 / (57.362 - sigma)
    return wl / f


def calculate_n(wl):
    """
    Calculate *n*, the refractive index of light at a given wavelength.

    :param wl: input wavelength (in vacuum)
    :type wl: np.array

    :return: numpy.ndarray
    """
    wl = np.asarray(wl)

    sigma = (1e4 / wl) ** 2
    f = 1.0 + 0.05792105 / (238.0185 - sigma) + 0.00167917 / (57.362 - sigma)
    new_wl = wl / f
    n = wl / new_wl
    print(n)


def vacuum_to_air_SLOAN(wl):
    """
    Converts vacuum wavelengths to air wavelengths using the outdated SLOAN definition.
    From the SLOAN website:

    AIR = VAC / (1.0 + 2.735182E-4 + 131.4182 / VAC^2 + 2.76249E8 / VAC^4)

    :param wl:
        The input wavelengths to convert


    """
    wl = np.asarray(wl)
    air = wl / (1.0 + 2.735182e-4 + 131.4182 / wl**2 + 2.76249e8 / wl**4)
    return air


def air_to_vacuum(wl):
    """
    Convert air wavelengths to vacuum wavelengths.

    :param wl: input air wavelegths
    :type wl: np.array

    :return: numpy.ndarray

    .. warning::

        It is generally not recommended to do this, as the function is imprecise.
    """
    wl = np.asarray(wl)

    sigma = 1e4 / wl
    vac = wl + wl * (
        6.4328e-5 + 2.94981e-2 / (146 - sigma**2) + 2.5540e-4 / (41 - sigma**2)
    )
    return vac


@np.vectorize
def idl_float(idl_num: str) -> float:
    """
    Convert an IDL string number in scientific notation to a float

    Parameters
    ----------
    idl_num : str
        Input str

    Returns
    -------
    float
        Output float

    Examples
    --------
    ```python
    >>> idl_float("1.6D4")
    1.6e4
    ```
    """
    idl_str = idl_num.lower()
    return float(idl_str.replace("d", "e"))



import os

import matplotlib.pyplot as plt
from matplotlib import gridspec
import numpy as np

plt.style.use("seaborn")


def plot_reconstructed(emulator, grid, folder):
    """
    Plot the reconstructed spectra and residual at each grid point.

    Parameters
    ----------
    emulator : :class:`Emulaor`
    grid : :class:`Starfish.grid_tools.GridInterface`
    folder : str or path-like

    Example of a reconstructed spectrum at [6200, 4.5, 0.5]

    .. figure:: assets/6200_4.50_0.00.png
        :align: center
        :width: 80%
    """
    recon_fluxes = []
    recon_err = []
    for params in grid.grid_points:
        flux, var = emulator.load_flux(params, full_cov=False)
        recon_fluxes.append(flux)
        recon_err.append(np.sqrt(var))
    fluxes = np.array(list(grid.fluxes))
    data = zip(grid.grid_points, fluxes, recon_fluxes, recon_err)

    plotdir = os.path.expandvars(folder)
    if not os.path.exists(plotdir):
        os.mkdir(plotdir)

    # Define the plotting function
    def plot(datum):
        par, real, recon, err = datum
        fig, ax = plt.subplots(nrows=2, figsize=(8, 8))
        ax[0].plot(emulator.wl, real, lw=1, label="Original")
        ax[0].plot(emulator.wl, recon, lw=1, label="Reconstructed")
        ax[0].fill_between(
            emulator.wl, recon - 2 * err, recon + 2 * err, color="C1", alpha=0.4
        )
        ax[0].set_ylabel(r"$f_\lambda$ [erg/cm^2/s/A]")
        ax[0].legend()

        ax[1].plot(emulator.wl, real - recon, lw=1, label="Residuals")
        ax[1].set_xlabel(r"$\lambda$ [A]")
        ax[1].set_ylabel(r"$\Delta f_\lambda$")
        ax[1].legend()

        name = "T={} logg={} Z={}".format(*par)
        ax[0].set_title(name)
        plt.tight_layout()
        filename = os.path.join(plotdir, "{:.0f}_{}_{}.png".format(*par))
        plt.savefig(filename)
        plt.close()

    list(map(plot, data))


def plot_eigenspectra(emulator, params, filename=None):
    """

    Parameters
    ----------
    emulator
    params
    filename : str or path-like, optional
        If provided, will save the plot at the given filename

    Example of a deconstructed set of eigenspectra

    .. figure:: assets/eigenspectra.png
        :align: center
    """
    weights = emulator.draw_weights(params)
    reconstructed = weights @ emulator.eigenspectra

    height = int(emulator.ncomps) * 1.25
    fig = plt.figure(figsize=(8, height))
    gs = gridspec.GridSpec(
        int(emulator.ncomps) + 1,
        1,
        height_ratios=[3] + list(np.ones(int(emulator.ncomps))),
    )
    ax = plt.subplot(gs[0])
    ax.plot(emulator.wl, reconstructed, lw=1)
    ax.set_ylabel("$f_\lambda$ [erg/cm^2/s/A]")
    plt.setp(ax.get_xticklabels(), visible=False)
    for i in range(emulator.ncomps):
        ax = plt.subplot(gs[i + 1], sharex=ax)
        ax.plot(emulator.wl, emulator.eigenspectra[i], c="0.4", lw=1)
        ax.set_ylabel(rf"$\xi_{i}$")
        if i < emulator.ncomps - 1:
            plt.setp(ax.get_xticklabels(), visible=False)
        ax.legend([rf"$w_{i}$ = {weights[i]:.2e}"])
    plt.xlabel("Wavelength (A)")
    plt.tight_layout(h_pad=0.2)

    plt.show()

    if filename is not None:
        plt.savefig(os.path.expandvars(filename))


def plot_emulator(emulator):
    import itertools

    T = np.unique(emulator.grid_points[:, 0])
    logg = np.unique(emulator.grid_points[:, 1])
    Z = np.unique(emulator.grid_points[:, 2])
    params = np.array(list(itertools.product(T, logg[:1], Z[:1])))
    idxs = np.array([emulator.get_index(p) for p in params])
    weights = emulator.weights[idxs.astype("int")].T
    if emulator.ncomps < 4:
        fix, axes = plt.subplots(
            emulator.ncomps, 1, sharex=True, figsize=(8, (emulator.ncomps - 1) * 2)
        )
    else:
        fix, axes = plt.subplots(
            int(np.ceil(emulator.ncomps / 2)),
            2,
            sharex=True,
            figsize=(13, (emulator.ncomps - 1) * 2),
        )
    axes = np.ravel(np.array(axes).T)
    [ax.set_ylabel(f"$w_{i}$") for i, ax in enumerate(axes)]
    for i, w in enumerate(weights):
        axes[i].plot(T, w, "o")

    Ttest = np.linspace(T.min(), T.max(), 100)
    Xtest = np.array(list(itertools.product(Ttest, logg[:1], Z[:1])))
    mus = []
    covs = []
    for X in Xtest:
        m, c = emulator(X)
        mus.append(m)
        covs.append(c)
    mus = np.array(mus)
    covs = np.array(covs)
    sigs = np.sqrt(np.diagonal(covs, axis1=-2, axis2=-1))
    for i, (m, s) in enumerate(zip(mus.T, sigs.T)):
        axes[i].plot(Ttest, m, "C1")
        axes[i].fill_between(Ttest, m - 2 * s, m + 2 * s, color="C1", alpha=0.4)
    axes[-1].set_xlabel("T (K)")
    plt.suptitle(
        f"Weights for fixed $\log g={logg[0]:.2f}$, $[Fe/H]={Z[0]:.2f}$", fontsize=20
    )
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
